from operator import truediv
import uuid
import json
import datetime
from arches.app.functions.base import BaseFunction
from arches.app.models import models
from arches.app.models.tile import Tile
from arches.app.datatypes.datatypes import DataTypeFactory
import logging


details = {
    'name': 'Consultation Status',
    'type': 'node',
    'description': 'Triggers change in status node of consultation instance',
    'defaultconfig': {
        "cons_comp_date_nodeid":"40eff4ce-893a-11ea-ae2e-f875a44e0e11",
        "cons_status_bool_nodeid":"6a773228-db20-11e9-b6dd-784f435179ea",
        },
    'classname': 'ConsultationStatusFunction',
    'component': '',
    'functionid':'96efa95a-1e2c-4562-ac1f-b415796f9f75'
}

class ConsultationStatusFunction(BaseFunction): 
    
    def handle_boolean_tile(self,tile,nodegroupid,nodeid,status):
        try:
            nodeid_str = str(nodeid)
            cons_status_tile = None

            cons_status_tile_filter = Tile.objects.filter(nodegroup_id=nodegroupid, resourceinstance_id=tile.resourceinstance_id)
            if len(cons_status_tile_filter) > 0:
                cons_status_tile = cons_status_tile_filter[0]
            else:
                cons_status_tile = Tile().get_blank_tile_from_nodegroup_id(resourceid = tile.resourceinstance_id,nodegroup_id = nodegroupid)

            if cons_status_tile != None:

                cons_status_tile_data = cons_status_tile.data
                cons_status_tile_data[nodeid_str] = status
                try:
                    cons_status_tile.save()
                    return True

                except Exception as e:
                    self.logger.error(e, "Could not save boolean tile")
                    return False

            else:
                return False

        except Exception as e:
                self.logger.error(e, "Could not create boolean tile")
                return False


    def save_cons_tile(self, tile, request, is_function_save_method=True):

        if request is None and is_function_save_method == True:
            return

        try:
            cons_status_bool_nodeid = self.config["cons_status_bool_nodeid"]
            cons_status_bool_filter = Tile.objects.filter(resourceinstance_id=tile.resourceinstance_id,nodegroup_id=cons_status_bool_nodeid)
            date_comp_node = models.Node.objects.get(nodeid=self.config["cons_comp_date_nodeid"])
            if len(cons_status_bool_filter) > 0:
                date_comp_tile = None
                if tile.nodegroup_id == str(date_comp_node.nodegroup_id):
                        date_comp_tile = tile
                else:
                    date_comp_tile_filter = Tile.objects.filter(resourceinstance_id=tile.resourceinstance_id,nodegroup_id=date_comp_node.nodegroup_id)
                    if len(date_comp_tile_filter) > 0:
                        date_comp_tile = date_comp_tile_filter[0]
                if  (date_comp_tile != None) and (tile.nodegroup_id == str(date_comp_tile.nodegroup_id)):
                    datatype_factory = DataTypeFactory()
                    datatype = datatype_factory.get_instance(date_comp_node.datatype)
                    date_comp_value = datatype.get_display_value(tile,date_comp_node)
                    if date_comp_value != None:
                        try:
                            self.handle_boolean_tile(tile,cons_status_bool_nodeid,cons_status_bool_nodeid,False)
                            return
                        except Exception as e:
                            self.logger.error(e,"Could not handle boolean tile after confirmation completion date node is present")
                            return
                    else:
                        return

            else:
                self.handle_boolean_tile(tile,cons_status_bool_nodeid,cons_status_bool_nodeid,True)
                return

        except Exception as e:
            self.logger.error(e,"Could not check boolean tile")
            return

    def save(self, tile, request, context=None):
        self.logger = logging.getLogger(__name__)
        try:
            self.save_cons_tile(tile=tile, request=request, is_function_save_method=True)
            return
        except Exception as e:
            self.logger.error(e, "Could not save tile")

    
    def delete(self,tile,request):
        self.logger = logging.getLogger(__name__)
        try:
            date_comp_node = models.Node.objects.get(nodeid=self.config["cons_comp_date_nodeid"])
            cons_status_bool_nodeid = self.config["cons_status_bool_nodeid"]
            if tile.nodegroup_id == date_comp_node.nodegroup_id:
                cons_status_bool_filter = Tile.objects.filter(resourceinstance_id=tile.resourceinstance_id,nodegroup_id=cons_status_bool_nodeid)
                if len(cons_status_bool_filter) > 0:
                        self.handle_boolean_tile(tile,cons_status_bool_nodeid,cons_status_bool_nodeid,True)
                        return
                else:
                    return
            else:
                return
        except Exception as e:
                self.logger.error(e, "Could not delete tile")
    
    def on_import(self,tile):
        raise NotImplementedError

    
    def after_function_save(self,tile,request):
        raise NotImplementedError
    
    
    def get(self):
        raise NotImplementedError



from django.core.management.base import BaseCommand
from arches.app.models.system_settings import settings
from pathlib import Path
from docx import Document
import logging

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "operation",
            nargs="?",
            choices=["fix_style_runs"],
            help="Fix broken DOCX style runs",
        )
        parser.add_argument(
            "-d",
            "--dest_dir",
            action="store",
            dest="dest_dir",
            default="docx",
            help="Destination directory for DOCX files, e.g. 'letter/templates'; default is 'docx'",
        )

    def handle(self, *args, **options):
        if options["operation"] == "fix_style_runs":
            self.fix_style_runs(**options)

    def fix_style_runs(self, **options):
        start_folder = Path(settings.ROOT_DIR).parent.parent
        docx_files = list(start_folder.glob(f"**/{options['dest_dir']}/*.docx"))
        if not docx_files:
            message = f"No DOCX files found in {options['dest_dir']} directory"
            print(message)
            logger.info(message)
            return
        for filename in docx_files:
            document = self.open_document(filename)
            if not document:
                continue

            save_required = self.process_paragraphs(document)
            self.save_document(document, filename, save_required)

    def open_document(self, filename):
        try:
            return Document(filename)
        except Exception as e:
            message = f"Error opening {filename}: {e}"
            print(message)
            logger.error(message)
            return None

    def process_paragraphs(self, document):
        save_required = False

        for p in document.paragraphs:
            run_text, start_pos, end_pos = self.collect_run_text(p)
            if start_pos + end_pos > 0:
                save_required |= self.replace_runs(p, run_text, start_pos, end_pos)

        return save_required

    def collect_run_text(self, paragraph):
        run_text = ""
        capture = False
        start_pos = end_pos = -1

        for i, r in enumerate(paragraph.runs):
            text = r.text
            if text.startswith("<") and text.endswith(">"):
                break
            if "<" in text:
                capture = True
                start_pos = i
            if capture:
                run_text += text
            if ">" in text:
                capture = False
                end_pos = i

        return run_text, start_pos, end_pos

    def replace_runs(self, paragraph, run_text, start_pos, end_pos):
        save_required = False
        for i, r in enumerate(paragraph.runs):
            if i < start_pos:
                continue
            if i == start_pos:
                if r.text != run_text:
                    r.text = run_text
                    save_required = True
            elif i <= end_pos:
                r.text = ""
                save_required = True

        return save_required

    def save_document(self, document, filename, save_required):
        if save_required:
            try:
                document.save(filename)
                message = f"DOCX style runs fixed in {filename}"
                print(message)
                logger.info(message)
            except Exception as e:
                message = f"Error saving {filename}: {e}"
                print(message)
                logger.error(message)
        else:
            message = f"No broken DOCX style runs found in {filename}"
            print(message)
            logger.info(message)



import os
from django.core.exceptions import ImproperlyConfigured
import ast


def get_env_variable(var_name):
    msg = "Set the %s environment variable"
    try:
        return os.environ[var_name]
    except KeyError:
        error_msg = msg % var_name
        raise ImproperlyConfigured(error_msg)


def get_optional_env_variable(var_name):
    try:
        return os.environ[var_name]
    except KeyError:
        return None


# options are either "PROD" or "DEV" (installing with Dev mode set gets you extra dependencies)
MODE = get_env_variable("DJANGO_MODE")

DEBUG = ast.literal_eval(get_env_variable("DJANGO_DEBUG"))

COUCHDB_URL = "http://{}:{}@{}:{}".format(
    get_env_variable("COUCHDB_USER"), get_env_variable("COUCHDB_PASS"), get_env_variable("COUCHDB_HOST"), get_env_variable("COUCHDB_PORT")
)  # defaults to localhost:5984

DATABASES = {
    "default": {
        "ENGINE": "django.contrib.gis.db.backends.postgis",
        "NAME": get_env_variable("PGDBNAME"),
        "USER": get_env_variable("PGUSERNAME"),
        "PASSWORD": get_env_variable("PGPASSWORD"),
        "HOST": get_env_variable("PGHOST"),
        "PORT": get_env_variable("PGPORT"),
        "POSTGIS_TEMPLATE": "template_postgis",
    }
}

CELERY_BROKER_URL = "amqp://{}:{}@arches_rabbitmq".format(
    get_env_variable("RABBITMQ_USER"), get_env_variable("RABBITMQ_PASS")
)  # RabbitMQ --> "amqp://guest:guest@localhost",  Redis --> "redis://localhost:6379/0"

# CANTALOUPE_HTTP_ENDPOINT = "http://{}:{}".format(get_env_variable("CANTALOUPE_HOST"), get_env_variable("CANTALOUPE_PORT"))
ELASTICSEARCH_HTTP_PORT = get_env_variable("ESPORT")
ELASTICSEARCH_HOSTS = [{"host": get_env_variable("ESHOST"), "port": ELASTICSEARCH_HTTP_PORT}]

USER_ELASTICSEARCH_PREFIX = get_optional_env_variable("ELASTICSEARCH_PREFIX")
if USER_ELASTICSEARCH_PREFIX:
    ELASTICSEARCH_PREFIX = USER_ELASTICSEARCH_PREFIX

ALLOWED_HOSTS = get_env_variable("DOMAIN_NAMES").split()

USER_SECRET_KEY = get_optional_env_variable("DJANGO_SECRET_KEY")
if USER_SECRET_KEY:
    # Make this unique, and don't share it with anybody.
    SECRET_KEY = USER_SECRET_KEY

STATIC_ROOT = "/static_root"



"""
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

from arches.app.models.system_settings import settings

def project_settings(request):
    return {
        "project_settings": {
            "APP_PATHNAME": settings.APP_PATHNAME
        }
    }



from django.conf import settings
from django.conf.urls.static import static
from django.conf.urls.i18n import i18n_patterns
from django.urls import include, path

# COPIED FROM ./arches_her/docker/aher_project/docker/urls.py

urlpatterns = [
    path('', include('arches.urls')),
   path("", include("arches_her.urls")),
] + static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)

# if settings.SHOW_LANGUAGE_SWITCH is True:
#     urlpatterns = i18n_patterns(*urlpatterns)



'''
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
'''

from arches.app.models import models
from arches.app.models.system_settings import settings
from arches_her.settings import APP_TITLE
from arches.app.models.resource import Resource
from arches.app.utils.betterJSONSerializer import JSONSerializer, JSONDeserializer
from django.shortcuts import render
from django.views.generic import TemplateView
from arches.app.datatypes.datatypes import DataTypeFactory
from arches.app.utils.permission_backend import get_createable_resource_types


class IndexView(TemplateView):

    template_name = ''

    def get(self, request):
        context = {}
        context['system_settings_graphid'] = settings.SYSTEM_SETTINGS_RESOURCE_MODEL_ID
        context['graph_models'] = []
        context['graphs'] = '[]'
        context['app_title'] = APP_TITLE
        context['plugins'] = []
        context['plugin_labels'] = {
            'active-consultations':'Active',
            'init-workflow':'New'
        }
        context['main_script'] = 'index'
        user_check = request.user.is_authenticated and request.user.username != 'anonymous'
        for plugin in models.Plugin.objects.all().order_by('sortorder'):
            if plugin.slug in context['plugin_labels'].keys() and request.user.has_perm('view_plugin', plugin) and user_check:
                plugin.name = context['plugin_labels'][plugin.slug]
                context['plugins'].append(plugin)

        context['user_is_reviewer'] = request.user.groups.filter(name='Resource Reviewer').exists()

        return render(request, 'index.htm', context)



# coding: utf-8
"""
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import json
import os
import re
import uuid
from datetime import datetime
import docx
import textwrap
from docx import Document
from docx.text.paragraph import Paragraph
from docx.oxml.parser import OxmlElement
from html.parser import HTMLParser
from html.entities import name2codepoint
from django.core.files.uploadedfile import UploadedFile
from django.http import HttpRequest, HttpResponseNotFound
from django.utils.translation import gettext as _
from django.views.generic import View
from arches.app.datatypes.datatypes import DataTypeFactory
from arches.app.models import models
from arches.app.models.concept import ConceptValue
from arches.app.models.resource import Resource
from arches.app.models.system_settings import settings
from arches.app.models.tile import Tile
from arches.app.utils.response import JSONResponse
from arches.app.views.tile import TileData
import site


class FileTemplateView(View):
    def __init__(self):
        self.doc = None
        self.resource = None

    def get(self, request):
        parenttile_id = request.GET.get("parenttile_id")
        parent_tile = Tile.objects.get(tileid=parenttile_id)
        letter_tiles = Tile.objects.filter(parenttile=parent_tile)
        file_list_node_id = "96f8830a-8490-11ea-9aba-f875a44e0e11"  # Digital Object
        url = None
        for tile in letter_tiles:
            if url is not None:
                break
            for data_obj in tile.data[file_list_node_id]:
                if data_obj["status"] == "uploaded":
                    url = data_obj["url"]
                    break

        if url is not None:
            return JSONResponse({"msg": "success", "download": url})
        return HttpResponseNotFound("No letters tile matching query by parent tile")

    def post(self, request):
        datatype_factory = DataTypeFactory()
        template_id = request.POST.get("template_id")
        parenttile_id = request.POST.get("parenttile_id")
        resourceinstance_id = request.POST.get("resourceinstance_id", None)
        transaction_id = request.POST.get("transaction_id", uuid.uuid1())
        self.resource = Resource.objects.get(resourceinstanceid=resourceinstance_id)
        self.resource.load_tiles()

        template_name = self.get_template_path(template_id)
        filename, file_extension = os.path.splitext(template_name)
        template_path = os.path.join(site.getsitepackages()[0], "arches_her", "docx", template_name)
        if not os.path.exists(template_path):
            template_path = os.path.join(settings.HER_ROOT, "docx", template_name)

        uploaded_docx_path = os.path.join(settings.APP_ROOT, "uploadedfiles", "docx")
        if not os.path.exists(uploaded_docx_path):
            os.mkdir(uploaded_docx_path)

        try:
            self.doc = Document(template_path)
        except FileNotFoundError:
            return HttpResponseNotFound("No Template Found")

        self.edit_letter(self.resource, datatype_factory)

        current_datetime = datetime.today()
        date = current_datetime.strftime("%Y-%m-%d")
        time = current_datetime.strftime("%H%M%S%f")

        new_file_name = f"{date}_{filename}_{time}{file_extension}"
        new_file_path = os.path.join(settings.APP_ROOT, "uploadedfiles", "docx", new_file_name)

        new_req = HttpRequest()
        new_req.method = "POST"
        new_req.user = request.user
        new_req.POST["data"] = None
        host = request.get_host()

        self.doc.save(new_file_path)
        saved_file = open(new_file_path, "rb")
        stat = os.stat(new_file_path)
        file_data = UploadedFile(saved_file)
        file_list_node_id = "96f8830a-8490-11ea-9aba-f875a44e0e11"  # Digital Object

        tile = json.dumps(
            {
                "tileid": None,
                "data": {
                    file_list_node_id: [
                        {
                            "name": new_file_name,
                            "accepted": True,
                            "height": 0,
                            "lastModified": stat.st_mtime,
                            "size": stat.st_size,
                            "status": "queued",
                            "type": "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
                            "width": 0,
                            "url": None,
                            "file_id": None,
                            "index": 0,
                            "content": "blob:" + host + "/{0}".format(uuid.uuid4()),
                        }
                    ]
                },
                "nodegroup_id": "7db68c6c-8490-11ea-a543-f875a44e0e11",
                "parenttile_id": None,
                "resourceinstance_id": "",
                "sortorder": 0,
                "tiles": {},
            }
        )

        new_req = HttpRequest()
        new_req.method = "POST"
        new_req.user = request.user
        new_req.POST["data"] = tile
        new_req.POST["transaction_id"] = transaction_id
        new_req.FILES["file-list_" + file_list_node_id] = file_data
        new_tile = TileData()
        new_tile.action = "update_tile"

        response = TileData.post(new_tile, new_req)
        if response.status_code == 200:
            tile = json.loads(response.content)
            return JSONResponse({"tile": tile, "status": "success"})

        return HttpResponseNotFound(response.status_code)

    def get_template_path(self, template_id):
        template_dict = {  # keys are valueids from "Letters" concept list; values are known file names
            "01dec356-e72e-40e6-b1b1-b847b9799d2f": "No progression letter.docx",  # Letter A
            "320abc26-db82-44a6-be11-8d44aaa23365": "No Need to Consult letter.docx",  # Letter A2
            "fd15c6c7-e94d-4914-8d51-a98bda6f4a7b": "Pre-app Predetermination letter.docx",  # Letter B1
            "8cc91474-11ce-47d9-b886-f0e3fc49d277": "Predetermination Letter.docx",  # Letter B2
            "08bb630d-a27b-45bc-a13f-567b428018c5": "Condition Two-Stage Letter.docx",  # Letter C
            "e14bd058-e9f2-48f8-8ef5-337310c3420f": "Pre-App Recommend Condition Letter.docx",  # Letter D1
            "92e745c3-7157-4831-bce9-73792d32abec": "Condition Investigation Letter.docx",  # Letter D2
            "41f3d0bb-a94d-469f-80c8-85ab03283972": "Condition Historic Building Recording Letter.docx",  # Letter D3
            "7f1e7061-8bb0-4338-9342-118f1e9214aa": "WSI Approval Letter.docx",  # Letter F1
            "eaa8a075-50e6-4c3d-ac08-fbe84865f577": "WSI Amend Letter.docx",  # Letter F2
            "8d605e5c-d0da-4b72-9ce3-2f7dac3381d1": "Post Excavation Assessment Approval Letter.docx",  # Letter G - PXA Approval
            "a31061ea-9b80-435f-82c8-94dc10afcbae": "Condition Satisfied Letter.docx",  # Letter H
            "eed24dd2-85a0-4402-a6ba-bda426b5da89": "Blank Adviser Letter.docx",  # Letter I - Bespoke Letter
            # No template available yet
            "a26c77ff-1d04-4b76-a45f-417f7ed24333": "",  # Additional Condition Text
            "8c12a812-8000-4ec9-913d-c6fd516117f2": "",  # Archaeological Recommendation Text
            # No concept selection available
            "missing 0": "Conditions Scope Notes.docx",
            "missing 1": "Mitigation Scope Notes.docx",
        }
        for key, value in list(template_dict.items()):
            if key == template_id:
                return value

        return None

    def edit_letter(self, consultation, datatype_factory):
        template_dict = {
            "Reference": "c128bf36-9384-11ea-bfe1-f875a44e0e11",
            "Primary Reference Number": "b37552bd-9527-11ea-97f4-f875a44e0e11",
            "Casework Officer": "4ea4a197-184f-11eb-9152-f875a44e0e11",
            "Completion Date": "40eff4ce-893a-11ea-ae2e-f875a44e0e11",
            "Consultation Name": "4ad69684-951f-11ea-b5c3-f875a44e0e11",
            "Proposal Description": "1b0e15ec-8864-11ea-8493-f875a44e0e11",
            "Log Date": "40eff4cd-893a-11ea-b0cc-f875a44e0e11",
            "Signature": "4ea4a197-184f-11eb-9152-f875a44e0e11",
            "Archaeological Priority Area": "58a2b98f-a255-11e9-9a30-00224800b26d",
            "Assessment of Significance": "8d41e4e0-a250-11e9-877f-00224800b26d",
            "Condition Type": "56fa335d-06fa-11eb-8328-f875a44e0e11",
            "Condition": "c36808b0-952c-11ea-9ff0-f875a44e0e11",
            "Mitigation Type": "e2585f8a-51a3-11eb-a7be-f875a44e0e11",
            "Mitigation": "bfd39106-51a3-11eb-9104-f875a44e0e11",
        }

        self.replace_in_letter(consultation.tiles, template_dict, datatype_factory)

    def replace_in_letter(self, tiles, template_dict, datatype_factory):
        mapping_dict = {
            "Reference": "",
            "Primary Reference Number": "",
            "Casework Officer": "",
            "Completion Date": "",
            "Consultation Name": "",
            "Proposal Description": "",
            "Log Date": "",
            "Signature": "",
            "Archaeological Priority Area": "",
            "Assessment of Significance": "",
            "Condition": "",
            "Mitigation": "",
            "Casework Officer Email": "",
            "Casework Officer Number": "",
            "Contact Name": "",
            "Address of consulting organisation": "",
            "Name of person consulting": "",
        }

        def get_value_from_tile(tile, node_id):
            current_node = models.Node.objects.get(nodeid=node_id)
            datatype = datatype_factory.get_instance(current_node.datatype)
            returnvalue = datatype.get_display_value(tile, current_node)
            return "" if returnvalue is None else returnvalue
        
        def remove_non_xml_compatible_chars(s: str) -> str:
            """
            Remove characters that are not compatible with XML.
            XML 1.0 compatible characters:
            - U+0009, U+000A, U+000D (whitespace characters)
            - U+0020 to U+D7FF
            - U+E000 to U+FFFD
            XML 1.1 additionally allows:
            - U+0001 to U+0008, U+000B to U+000C, U+000E to U+001F
            But for broad compatibility, we'll target XML 1.0 here.
            :param s: Input string
            :return: String with non-XML-compatible characters removed
            """
            # For XML 1.0, adjust ranges if targeting XML 1.1
            return "".join(
                char
                for char in s
                if ord(char) in (0x9, 0xA, 0xD)
                or 0x20 <= ord(char) <= 0xD7FF
                or 0xE000 <= ord(char) <= 0xFFFD
            )

        # Advice and Conditions.
        advice_nodegroup_id = "8d41e49f-a250-11e9-b6b3-00224800b26d"
        advice_node_id = "c36808b0-952c-11ea-9ff0-f875a44e0e11"
        advice_type_node_id = "56fa335d-06fa-11eb-8328-f875a44e0e11"
        conditions = []

        # Action and Mitigations.

        action_nodegroup_id = "a5e15f5c-51a3-11eb-b240-f875a44e0e11"
        action_node_id = "bfd39106-51a3-11eb-9104-f875a44e0e11"
        action_type_node_id = "e2585f8a-51a3-11eb-a7be-f875a44e0e11"
        mitgations_concept_id = "f60c394e-c99f-4c91-9f68-791465036cde"

        mitigations = []

        mitigation_scope_dict = {}

        concepts_from_mitigation_group = models.Relation.objects.filter(conceptfrom=mitgations_concept_id)
        for mitigation_concept in concepts_from_mitigation_group:
            mitigation_concept_to_value = models.Value.objects.filter(concept=mitigation_concept.conceptto_id)
            for mitigation_value in mitigation_concept_to_value:
                if str(mitigation_value.valuetype_id) == "prefLabel":
                    mitigation_scope_dict[mitigation_value.value] = str(mitigation_value.valueid)
                elif str(mitigation_value.valuetype_id) == "scopeNote":
                    value_id = models.Value.objects.filter(concept=mitigation_value.concept_id, valuetype="prefLabel")
                    mitigation_scope_dict[str(value_id[0].valueid)] = mitigation_value.value
                else:
                    pass

        for tile in tiles:
            mitigation = {}
            condition = {}
            if str(tile.nodegroup_id) == action_nodegroup_id:

                mitigation_scopenote = mitigation_scope_dict.get(
                    mitigation_scope_dict.get(get_value_from_tile(tile, action_type_node_id)), ""
                )

                # if len(mitigation_scopenote) > 0:
                #     mitigation_scopenote = "<i>" + mitigation_scopenote + "</i>"
                insert_break = len(mitigation_scopenote) > 0
                mitigation[
                    "content"
                ] = f"{'<br>' if insert_break else ''}{mitigation_scopenote}{'<br>' if insert_break else ''}{get_value_from_tile(tile, action_node_id)}"
                mitigation["type"] = get_value_from_tile(tile, action_type_node_id)
            elif str(tile.nodegroup_id) == advice_nodegroup_id:
                condition["content"] = get_value_from_tile(tile, advice_node_id)
                template_name = self.get_template_path(self.request._post["template_id"])
                if template_name == "WSI Amend Letter.docx" or template_name == "WSI Approval Letter.docx":
                    condition["type"] = ""
                else:
                    condition["type"] = f"{get_value_from_tile(tile, advice_type_node_id)}"
            else:
                for key, value in list(template_dict.items()):
                    if value in tile.data:
                        lookup_val = get_value_from_tile(tile, value)
                        try:
                            mapping_dict[key] = lookup_val
                        except TypeError:
                            pass
            if len(mitigation) > 0:
                mitigations.append(mitigation)
            elif len(condition) > 0:
                conditions.append(condition)

            contactNodeId = "b7304f4c-3ace-11eb-8884-f875a44e0e11"
            contacts = {
                "Applicant": "4ea4a19a-184f-11eb-aef8-f875a44e0e11",
                "Planning Officer": "4ea4a192-184f-11eb-a0d6-f875a44e0e11",
                "Owner": "4ea4c885-184f-11eb-b4d5-f875a44e0e11",
                "Agent": "4ea4c884-184f-11eb-b64d-f875a44e0e11",
                "Casework Officer": "4ea4a197-184f-11eb-9152-f875a44e0e11",
            }

            addressNodegroupId = "5f93048e-80a9-11ea-b0da-f875a44e0e11"
            nameNodegroupId = "4110f741-1a44-11e9-885e-000d3ab1e588"
            contactDetailsNodegroupId = "2547c12f-9505-11ea-a507-f875a44e0e11"
            contactNameForCorrespondenceNodeId = "2beefb56-4084-11eb-bcc5-f875a44e0e11"
            fullnameNodeId = "5f8ded26-7ef9-11ea-8e29-f875a44e0e11"
            firstnameNodeId = "2caeb5e7-7b44-11ea-a919-f875a44e0e11"
            lastnameNodeId = "96a3942a-7e53-11ea-8b5a-f875a44e0e11"
            nameTitleNodeId = "6da2f03b-7e55-11ea-8fe5-f875a44e0e11"
            nameUseTypeNodeId = "4110f747-1a44-11e9-96b7-000d3ab1e588"
            forCorrespondenceNameValueId = "85c26c81-e356-4454-a2ba-67e7ad9b95cd"
            primaryNameValueId = "2df285fa-9cf2-45e7-bc05-a67b7d7ddc2f"
            contactPointNodeId = "2547c133-9505-11ea-8e49-f875a44e0e11"
            contactPointTypeNodeId = "2547c132-9505-11ea-b22f-f875a44e0e11"
            contactPointTypeMailValueId = "e6d433a2-7f77-4eb7-96f2-57ebe0ac251e"
            addressDict = {
                "Building Name": "b3a2761d-effb-11eb-9867-a87eeabdefba",
                "Building Number": "b3a2761f-effb-11eb-9059-a87eeabdefba",
                "Street": "b3a27621-effb-11eb-83e6-a87eeabdefba",
                "Locality": "b3a28c1a-effb-11eb-a811-a87eeabdefba",
                "Town or City": "b3a27617-effb-11eb-a80f-a87eeabdefba",
                "Postcode": "b3a27619-effb-11eb-a66d-a87eeabdefba",
            }
            if contactNodeId in tile.data:
                caseAgentResourceId = None
                contactResourceiId = None
                if tile.data[contacts["Casework Officer"]]:
                    caseAgentResourceId = tile.data[contacts["Casework Officer"]][0]["resourceId"]
                if caseAgentResourceId:
                    caseAgentResource = Resource.objects.get(resourceinstanceid=caseAgentResourceId)
                    caseAgentResource.load_tiles()
                    for caseAgentTile in caseAgentResource.tiles:
                        if caseAgentTile.nodegroup.nodegroupid == uuid.UUID(contactDetailsNodegroupId):
                            if caseAgentTile.data[contactPointTypeNodeId] == "0f466b8b-a347-439f-9b61-bee9811ccbf0":
                                mapping_dict["Casework Officer Email"] = get_value_from_tile(caseAgentTile, contactPointNodeId)
                            elif caseAgentTile.data[contactPointTypeNodeId] == "75e6cfad-7418-4ed3-841b-3c083d7df30b":
                                mapping_dict["Casework Officer Number"] = get_value_from_tile(caseAgentTile, contactPointNodeId)

                if tile.data[contactNodeId] == "5cc97bfd-d76f-40fc-be60-fbb9dfb28fc4" and tile.data[contacts["Planning Officer"]]:
                    contactResourceiId = tile.data[contacts["Planning Officer"]][0]["resourceId"]
                elif tile.data[contactNodeId] == "d88aa873-848c-45cb-b967-4febe7397912" and tile.data[contacts["Owner"]]:
                    contactResourceiId = tile.data[contacts["Owner"]][0]["resourceId"]
                elif tile.data[contactNodeId] == "dcaf8850-9cfc-44ea-9fd4-0ca419806e2b" and tile.data[contacts["Agent"]]:
                    contactResourceiId = tile.data[contacts["Agent"]][0]["resourceId"]

                if contactResourceiId:
                    contactResource = Resource.objects.get(resourceinstanceid=contactResourceiId)
                    contactResource.load_tiles()

                    for contactTile in contactResource.tiles:
                        if contactTile.nodegroup.nodegroupid == uuid.UUID(nameNodegroupId):
                            if mapping_dict["Name of person consulting"] == "" or contactTile.data[nameUseTypeNodeId] == primaryNameValueId:
                                nameTitle = ConceptValue(contactTile.data[nameTitleNodeId]).value
                                fullName = "{0} {1}".format(
                                    get_value_from_tile(contactTile, firstnameNodeId), get_value_from_tile(contactTile, lastnameNodeId)
                                )
                                mapping_dict["Name of person consulting"] = "{0} {1}".format(nameTitle, fullName) if nameTitle else fullName
                        elif contactTile.nodegroup.nodegroupid == uuid.UUID(contactDetailsNodegroupId):
                            if contactTile.data[contactPointTypeNodeId] == contactPointTypeMailValueId:
                                mapping_dict["Contact Name"] = get_value_from_tile(contactTile, contactNameForCorrespondenceNodeId)
                                addressConsult = (
                                    get_value_from_tile(contactTile, contactPointNodeId).replace(", ", "<br>").replace(",", "<br>")
                                )
                                mapping_dict["Address of consulting organisation"] = addressConsult

            mapping_dict["Casework Officer"] = re.sub("\[\d+\] ", "", mapping_dict["Casework Officer"])
            mapping_dict["Signature"] = mapping_dict["Casework Officer"]

        for mitigation in mitigations:
            add_break = len(mitigation["content"]) > 0
            mapping_dict[
                "Mitigation"
            ] += f'<br><b>{mitigation["type"]}</b>{"<br>"if add_break else ""}{mitigation["content"]}{"<br>" if add_break else ""}'

        for condition in conditions:
            mapping_dict["Condition"] += "<b>{}</b>{}<br>".format(condition["type"], condition["content"])

        associate_heritage = mapping_dict["Archaeological Priority Area"]
        if associate_heritage == "":
            mapping_dict["Archaeological Priority Area"] = "The planning application is not in an Archaeological Priority Area."
        else:
            mapping_dict[
                "Archaeological Priority Area"
            ] = "The planning application lies in an area of archaeological interest (Archaeological Priority Area) identified in the Local Plan: {}".format(
                associate_heritage
            )

        if mapping_dict["Assessment of Significance"] != "":
            mapping_dict["Assessment of Significance"] += "<br>"

        htmlTags = re.compile(r"<(?:\"[^\"]*\"['\"]*|'[^']*'['\"]*|[^'\">])+>")
        for key in mapping_dict:
            html = False
            if htmlTags.search(mapping_dict[key] if mapping_dict[key] is not None else ""):
                html = True
            xml_compatible_string = remove_non_xml_compatible_chars(mapping_dict[key])
            self.replace_string(self.doc, key, xml_compatible_string, html)
            
            

    def replace_string(self, document, key, v, html=False):
        # Note that the intent here is to preserve how things are styled in the docx
        # easiest way is to iterate through p.runs, not as fast as iterating through parent.paragraphs
        # advantage of the former is that replacing run.text preserves styling, replacing p.text does not

        def parse_html_to_docx(p, k, v):
            style = p.style
            if k in p.text:
                p.clear()
                document_html_parser = DocumentHTMLParser(p, document)
                document_html_parser.insert_into_paragraph_and_feed(v)

        def replace_in_runs(p_list, k, v):
            for paragraph in p_list:
                if html is True:
                    parse_html_to_docx(paragraph, k, v)
                for i, run in enumerate(paragraph.runs):
                    if k in run.text:  # now check if html
                        run_style = run.style
                        run.text = run.text.replace(k, v)
                    elif (
                        i == (len(paragraph.runs) - 1) and k in paragraph.text
                    ):  # backstop case: rogue text outside of run obj - must fix template
                        paragraph.text = paragraph.text.replace(k, v)

        def iterate_tables(t_list, k, v):
            for table in t_list:
                for row in table.rows:
                    for cell in row.cells:
                        replace_in_runs(cell.paragraphs, k, v)

        if v is not None and key is not None:
            k = "<" + key + ">"
            doc = document
            # some of these are probably unnecessary
            # foot_style = styles['Footer']
            # head_style = styles['Header']
            # t_style = None
            # p_style = None
            run_style = None

            if len(doc.paragraphs) > 0:
                replace_in_runs(doc.paragraphs, k, v)

            if len(doc.tables) > 0:
                iterate_tables(doc.tables, k, v)

            if len(doc.sections) > 0:
                for section in doc.sections:
                    replace_in_runs(section.footer.paragraphs, k, v)
                    iterate_tables(section.footer.tables, k, v)
                    replace_in_runs(section.header.paragraphs, k, v)
                    iterate_tables(section.header.tables, k, v)

    def insert_image(self, document, k, v, image_path=None, config=None):
        # going to need to write custom logic depending on how images should be placed/styled

        return True

    def insert_custom(self, document, k, v, config=None):
        # perhaps replaces {{custom_object}} with pre-determined text structure with custom style/format

        return True


class DocumentHTMLParser(HTMLParser):
    def __init__(self, paragraph, document):
        HTMLParser.__init__(self)
        self.document = document
        self.paragraph = paragraph
        self.table = None
        self.table_cols = 0
        self.table_rows = 0
        self.max_cols_reached = False
        self.td_cursor = False
        self.hyperlink = False
        self.list_style = "ul"
        self.ol_counter = 1
        self.run = self.paragraph.add_run()

    def insert_paragraph_after(self, paragraph, text=None, style=None):
        """Insert a new paragraph after the given paragraph."""
        new_p = OxmlElement("w:p")
        paragraph._p.addnext(new_p)
        new_para = Paragraph(new_p, paragraph._parent)
        if text:
            new_para.add_run(text)
        if style is not None:
            new_para.style = style
        return new_para

    def add_hyperlink(self, paragraph, url, text, color=None, underline=None):
        # This gets access to the document.xml.rels file and gets a new relation id value
        part = self.paragraph.part
        r_id = part.relate_to(url, docx.opc.constants.RELATIONSHIP_TYPE.HYPERLINK, is_external=True)

        # Create the w:hyperlink tag and add needed values
        hyperlink = docx.oxml.shared.OxmlElement("w:hyperlink")
        hyperlink.set(
            docx.oxml.shared.qn("r:id"),
            r_id,
        )

        # Create a w:r element
        new_run = docx.oxml.shared.OxmlElement("w:r")

        # Create a new w:rPr element
        rPr = docx.oxml.shared.OxmlElement("w:rPr")

        # Add color if it is given
        if not color is None:
            c = docx.oxml.shared.OxmlElement("w:color")
            c.set(docx.oxml.shared.qn("w:val"), color)
            rPr.append(c)  # #5384da ; rgb(83,132,218)

        # Remove underlining if it is requested
        if not underline:
            u = docx.oxml.shared.OxmlElement("w:u")
            u.set(docx.oxml.shared.qn("w:val"), "none")
            rPr.append(u)

        # Join all the xml elements together add add the required text to the w:r element
        new_run.append(rPr)
        new_run.text = text
        hyperlink.append(new_run)
        paragraph._p.append(hyperlink)

        return hyperlink

    def insert_into_paragraph_and_feed(self, html):
        html = html.replace("\n\n", "<br>")
        self.run = self.paragraph.add_run()
        self.feed(html)

    def handle_starttag(self, tag, attrs):
        self.run = self.paragraph.add_run()
        if tag == "i" or tag == "em":
            self.run.italic = True
        if tag == "b" or tag == "strong":
            self.run.bold = True
        if tag == "s":
            self.run.strike = True
        if tag == "u":
            self.run.underline = True
        if tag == "ol":
            self.list_style = "ol"
        if tag == "ul":
            self.list_style = "ul"
        if tag in ["br", "ul", "ol"]:
            self.run.add_break()
        if tag == "li":
            if self.list_style == "ul":
                self.run.add_text("● ")
            else:
                self.run.add_text(str(self.ol_counter) + ". ")
                self.ol_counter += 1
        if tag == "p":
            self.run.add_break()
            # self.run.add_break()
            # self.run.add_tab()
        if tag == "a":
            self.hyperlink = attrs[0][1]
        if tag == "table":
            self.table = self.document.add_table(self.table_rows, self.table_cols)
            self.table.autofit = True
        if tag == "tr":
            self.table_rows += 1
            self.table.add_row()
        if tag == "td":
            self.table_cols += 1
            if self.max_cols_reached is False:
                self.table.add_column(1)
            self.td_cursor = True

    def handle_endtag(self, tag):
        if tag in ["br", "li", "ul", "ol"]:
            self.run.add_break()
        self.run = self.paragraph.add_run()
        if tag == "ol":
            self.ol_counter = 1
        if tag == "table":
            tbl = self.table._tbl
            p = self.paragraph._p
            p.addnext(tbl)
            self.table = None
            self.table_cols = 0
            self.table_rows = 0
        if tag == "tr":
            self.table_cols = 0
            self.max_cols_reached = True
        if tag == "td":
            self.td_cursor = False

    def handle_data(self, data):
        if "&#39;" in data:
            data = data.replace("&#39;", "'")

        if self.hyperlink is not False:
            blue = docx.shared.RGBColor(83, 132, 218)
            color = blue.__str__()
            self.add_hyperlink(self.paragraph, self.hyperlink, data, color)
            self.hyperlink = False
        elif self.td_cursor is True:
            self.table.cell(self.table_rows - 1, self.table_cols - 1).add_paragraph(data)  # formatting?
        else:
            self.run.add_text(data)

    def handle_entityref(self, name):
        c = chr(name2codepoint[name])
        self.run.add_text(c)

    def handle_charref(self, name):
        if name.startswith("x"):
            c = chr(int(name[1:], 16))
        else:
            c = chr(int(name))
        self.run.add_text(c)



# Generated by Django 3.2.19 on 2023-07-13 22:08

from django.db import migrations, models
from django.utils.translation import gettext as _


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("models", "9945_file_thumbnail_bin_file_thumbnail_text"),
    ]

    def add_plugins(apps, schema_editor):
        Plugin = apps.get_model("models", "Plugin")

        # Add Accessibility
        Plugin.objects.update_or_create(
            pluginid="8688eaf9-3605-4384-823c-d01c0d210f82",
            name="Accessibility",
            icon="fa fa-universal-access",
            component="views/components/plugins/accessibility",
            componentname="accessibility",
            slug="accessibility",
            config={},
            sortorder=0
        )

        Plugin.objects.update_or_create(
            pluginid="76a27df9-6f16-47ed-bd47-bb95c0fe7173",
            name="Active Consultations",
            icon="fa fa-compass",
            component="views/components/plugins/active-consultations",
            componentname="active-consultations",
            slug="active-consultations",
            config={},
            sortorder=0
            )

        Plugin.objects.update_or_create(
            pluginid="b2778828-a6ac-6481-c38b-fd463d878f1f",
            name="Application Area",
            icon="fa fa-building",
            component="views/components/plugins/application-area",
            componentname="application-area",
            config={"show":False},
            slug="application-area",
            sortorder=0
            )

        Plugin.objects.update_or_create(
            pluginid="4dc9bd5a-6e5c-440d-ae3c-af94396e2d72",
            name="Communication",
            icon="fa fa-comment",
            component="views/components/plugins/communication-workflow",
            componentname="communication-workflow",
            config={"show":False},
            slug="communication-workflow",
            sortorder=1
            )

        Plugin.objects.update_or_create(
            pluginid="a1667717-b7bd-4570-b27a-ec352c767e0e",
            name="Consultation",
            icon="fa fa-file",
            component="views/components/plugins/consultation-workflow",
            componentname="consultation-workflow",
            config={"show":False},
            slug="consultation-workflow",
            sortorder=0
            )

        Plugin.objects.update_or_create(
            pluginid="4bd762cf-b581-11e9-a7f9-784f435179ea",
            name="Correspondence",
            icon="fa fa-envelope",
            component="views/components/plugins/correspondence-workflow",
            componentname="correspondence-workflow",
            config={"show":False},
            slug="correspondence-workflow",
            sortorder=1
            )

        Plugin.objects.update_or_create(
            pluginid="0b1499e0-6cdc-403b-a2e3-499c2201069d",
            name="Site Visit",
            icon="fa fa-clipboard",
            component="views/components/plugins/site-visit",
            componentname="site-visit",
            config={"show":False},
            slug="site-visit",
            sortorder=0
        )

        Plugin.objects.update_or_create(
            pluginid="49507fb0-89c6-47b7-b506-9b2b29a3b8d8",
            name="Consultations",
            icon="fa fa-play-circle",
            component="views/components/plugins/init-workflow",
            componentname="init-workflow",
            config={
                "workflows": [
                    {
                        "workflowid": "b2778828-a6ac-6481-c38b-fd463d878f1f",
                        "slug":"application-area",
                        "name": "Application Area",
                        "icon": "fa fa-building",
                        "bgColor": "#87bceb",
                        "circleColor": "#a7cdf0",
                        "desc": "An area that may be re-developed or newly built"
                    },{
                        "workflowid": "a1667717-b7bd-4570-b27a-ec352c767e0e",
                        "slug":"consultation-workflow",
                        "name": "Consultation",
                        "icon": "fa fa-file",
                        "bgColor": "#62a3db",
                        "circleColor": "#8bbfea",
                        "desc": "Advice on the potential work related to an application area"
                    },{
                        "workflowid": "4dc9bd5a-6e5c-440d-ae3c-af94396e2d72",
                        "slug":"communication-workflow",
                        "name": "Communication",
                        "icon": "fa fa-comment",
                        "bgColor": "#9795EE",
                        "circleColor": "#8bbbe4",
                        "desc": "A conversation via phone, email, or other channel"
                    },{
                        "workflowid": "0b1499e0-6cdc-403b-a2e3-499c2201069d",
                        "slug":"site-visit",
                        "name": "Site Visit",
                        "icon": "fa fa-clipboard",
                        "bgColor": "#716FE0",
                        "circleColor": "#b0aff0",
                        "desc": "A visit to ensure that a condition has been met"
                    },{
                        "workflowid": "4bd762cf-b581-11e9-a7f9-784f435179ea",
                        "slug":"correspondence-workflow",
                        "name": "Correspondence",
                        "icon": "fa fa-envelope",
                        "bgColor": "#fed04f",
                        "circleColor": "#fedc7d",
                        "desc": "Letter or other document sent to an applicant"
                    }
                ],
                "show":True
            },
            slug="init-workflow",
            sortorder=0
            )


    def remove_plugins(apps, schema_editor):
        Plugin = apps.get_model("models", "Plugin")

        for plugin in Plugin.objects.filter(pk__in=[
            "8688eaf9-3605-4384-823c-d01c0d210f82",
            "76a27df9-6f16-47ed-bd47-bb95c0fe7173",
            "b2778828-a6ac-6481-c38b-fd463d878f1f",
            "4dc9bd5a-6e5c-440d-ae3c-af94396e2d72",
            "a1667717-b7bd-4570-b27a-ec352c767e0e",
            "4bd762cf-b581-11e9-a7f9-784f435179ea",
            "0b1499e0-6cdc-403b-a2e3-499c2201069d",
            "49507fb0-89c6-47b7-b506-9b2b29a3b8d8",
        ]):
            plugin.delete()

    def add_reports(apps, schema_editor):
        ReportTemplate = apps.get_model("models", "ReportTemplate")

        ReportTemplate.objects.update_or_create(
            name="Activity Template",
            component="views/components/reports/activity",
            defaultconfig={
                "uncompacted_reporting": True
            },
            description="Activity report",
            componentname="activity-report",
            templateid="eeb7054c-df37-45d1-8f09-39abb582077e",
            preload_resource_data=False
        )

        ReportTemplate.objects.update_or_create(
            name="Application Area Template",
            component="views/components/reports/application-area",
            defaultconfig={
                "uncompacted_reporting": True
            },
            description="Application Area report",
            componentname="application-area-report",
            templateid="9ed340ba-88c6-4982-9487-c3aae4b14d16",
            preload_resource_data=False
        )

        ReportTemplate.objects.update_or_create(
            name="Area Template",
            component="views/components/reports/area",
            defaultconfig={
                "uncompacted_reporting": True
            },
            description="Area report",
            componentname="area-report",
            templateid="c6f163ca-3ca5-4c50-b398-bee406fe059d",
            preload_resource_data=False
        )

        ReportTemplate.objects.update_or_create(
            name="Artefact Template",
            component="views/components/reports/artefact",
            defaultconfig={
                "uncompacted_reporting": True
            },
            description="Artefact report",
            componentname="artefact-report",
            templateid="b32fc2f2-5570-46c2-8d1e-09ca54c750ab",
            preload_resource_data=False
        )

        ReportTemplate.objects.update_or_create(
            name="Bibliographic Sourc3 Template",
            component="views/components/reports/bibliographic-source",
            defaultconfig={
                "uncompacted_reporting": True
            },
            description="Bibliographic Source report",
            componentname="bibliographic-source-report",
            templateid="44fdae23-7883-4d5c-88fa-7498b6fc898d",
            preload_resource_data=False
        )

        ReportTemplate.objects.update_or_create(
            name="Consultation Template",
            component="views/components/reports/consultation",
            defaultconfig={
                "uncompacted_reporting": True
            },
            description="Consultation report",
            componentname="consultation-report",
            templateid="465443ea-0e01-4895-a9cf-5144b02213da",
            preload_resource_data=False
        )

        ReportTemplate.objects.update_or_create(
            name="Digital Object Template",
            component="views/components/reports/digital-object",
            defaultconfig={
                "uncompacted_reporting": True
            },
            description="Digital Object report",
            componentname="digital-object-report",
            templateid="2bf88426-a053-4134-b006-d4dc2e120a6e",
            preload_resource_data=False
        )

        ReportTemplate.objects.update_or_create(
            name="Heritage Story Template",
            component="views/components/reports/heritage-story",
            defaultconfig={
                "uncompacted_reporting": True
            },
            description="Heritage Story report",
            componentname="heritage-story-report",
            templateid="6858ba4d-70a2-4e78-baf1-c827dd66956b",
            preload_resource_data=False
        )

        ReportTemplate.objects.update_or_create(
            name="Historic Aircraft Template",
            component="views/components/reports/historic-aircraft",
            defaultconfig={
                "uncompacted_reporting": True
            },
            description="Historic Aircraft report",
            componentname="historic-aircraft-report",
            templateid="655f86f0-643a-4df4-be8b-9e9030d2ff94",
            preload_resource_data=False
        )

        ReportTemplate.objects.update_or_create(
            name="Historic Landscape Characterization Template",
            component="views/components/reports/historic-landscape-characterization",
            defaultconfig={
                "uncompacted_reporting": True
            },
            description="Historic Landscape Characterization report",
            componentname="historic-landscape-characterization-report",
            templateid="5c757462-97d9-43fc-bd39-54367f6f3c2e",
            preload_resource_data=False
        )

        ReportTemplate.objects.update_or_create(
            name="Maritime Vessel Template",
            component="views/components/reports/maritime-vessel",
            defaultconfig={
                "uncompacted_reporting": True
            },
            description="Maritime Vessel report",
            componentname="maritime-vessel-report",
            templateid="03871d64-7076-46e8-b0ed-4b8813b0afd4",
            preload_resource_data=False
        )

        ReportTemplate.objects.update_or_create(
            name="Monument Template",
            component="views/components/reports/monument",
            defaultconfig={
                "uncompacted_reporting": True
            },
            description="Monument report",
            componentname="monument-report",
            templateid="523a5a7f-006b-4caf-a801-a1621967a88b",
            preload_resource_data=False
        )

        ReportTemplate.objects.update_or_create(
            name="Organization Template",
            component="views/components/reports/organization",
            defaultconfig={
                "uncompacted_reporting": True
            },
            description="Organization report",
            componentname="organization-report",
            templateid="407019fd-69e4-4aa4-a054-1466acaf8687",
            preload_resource_data=False
        )

        ReportTemplate.objects.update_or_create(
            name="Period Template",
            component="views/components/reports/period",
            defaultconfig={
                "uncompacted_reporting": True
            },
            description="Period report",
            componentname="period-report",
            templateid="4c2ed0b0-c278-49e2-9ddf-ba44aa77a6f3",
            preload_resource_data=False
        )

        ReportTemplate.objects.update_or_create(
            name="Person Template",
            component="views/components/reports/person",
            defaultconfig={
                "uncompacted_reporting": True
            },
            description="Person report",
            componentname="person-report",
            templateid="9e9f8e55-3ea9-4bd3-8b11-c1a519fbcfcc",
            preload_resource_data=False
        )

        ReportTemplate.objects.update_or_create(
            name="Place Template",
            component="views/components/reports/place",
            defaultconfig={
                "uncompacted_reporting": True
            },
            description="Place report",
            componentname="place-report",
            templateid="6b183add-a159-4d2b-a6b0-3f426d8c08a1",
            preload_resource_data=False
        )

    def remove_reports(apps, schema_editor):
        ReportTemplate = apps.get_model("models", "ReportTemplate")

        for report_template in ReportTemplate.objects.filter(pk__in=[
            "eeb7054c-df37-45d1-8f09-39abb582077e",
            "9ed340ba-88c6-4982-9487-c3aae4b14d16",
            "c6f163ca-3ca5-4c50-b398-bee406fe059d",
            "b32fc2f2-5570-46c2-8d1e-09ca54c750ab",
            "44fdae23-7883-4d5c-88fa-7498b6fc898d",
            "465443ea-0e01-4895-a9cf-5144b02213da",
            "2bf88426-a053-4134-b006-d4dc2e120a6e",
            "6858ba4d-70a2-4e78-baf1-c827dd66956b",
            "655f86f0-643a-4df4-be8b-9e9030d2ff94",
            "5c757462-97d9-43fc-bd39-54367f6f3c2e",
            "03871d64-7076-46e8-b0ed-4b8813b0afd4",
            "523a5a7f-006b-4caf-a801-a1621967a88b",
            "407019fd-69e4-4aa4-a054-1466acaf8687",
            "4c2ed0b0-c278-49e2-9ddf-ba44aa77a6f3",
            "9e9f8e55-3ea9-4bd3-8b11-c1a519fbcfcc",
            "6b183add-a159-4d2b-a6b0-3f426d8c08a1",
        ]):
            report_template.delete()


    def add_functions(apps, schema_editor):
        Function = apps.get_model("models", "Function")

        Function.objects.update_or_create(
            name="BNG Point to GeoJSON",
            functiontype="node",
            modulename="bngpoint_to_geojson_function.py",
            description="Pushes the geometry from a BNG Point node to a related GeoJSON node",
            defaultconfig={"bng_node": "", "geojson_node": "", "bng_nodegroup": "", "geojson_nodegroup": "", "triggering_nodegroups": []},
            classname="BNGPointToGeoJSON",
            component="views/components/functions/bngpoint-to-geojson-function",
            functionid="0434df8d-b98a-4b41-9a0a-68cd9214ad73",
        )

        Function.objects.update_or_create(    
            name="Consultation Status",
            functiontype="node",
            modulename="consultation_status_function.py",
            description="Triggers change in status node of consultation instance",
            defaultconfig={
                "cons_comp_date_nodeid":"40eff4ce-893a-11ea-ae2e-f875a44e0e11",
                "cons_status_bool_nodeid":"6a773228-db20-11e9-b6dd-784f435179ea",
                },
            classname="ConsultationStatusFunction",
            component="",
            functionid="96efa95a-1e2c-4562-ac1f-b415796f9f75"
        )

        Function.objects.update_or_create(    
            name="Generate Unique References",
            functiontype="node",
            modulename="generate_unique_references_function.py",
            description="Checks for Simple UID and Resource ID nodes populated and, if not populated, generates them",
            defaultconfig={
                "simpleuid_node": "",
                "resourceid_node": "",
                "uniqueresource_nodegroup": "",
                "triggering_nodegroups": [],
                "nodegroup_nodes": [],
            },
            classname="GenerateUniqueReferences",
            component="views/components/functions/generate-unique-references-function",
            functionid="39d627ae-6973-4ddb-8b62-1f0230e1e3f9",
        )
        
        Function.objects.update_or_create(
            name="GeoJSON to BNG Point",
            functiontype="node",
            modulename="geojson_to_bngpoint_function",
            description="Pushes the geometry from a GeoJSON node's centroid to a related BNG Point node",
            defaultconfig={
                "geojson_input_node": "",
                "bng_output_node": "",
                "geojson_input_nodegroup": "",
                "bng_output_nodegroup": "",
                "triggering_nodegroups": [],
            },
            classname="GeoJSONToBNGPoint",
            component="views/components/functions/geojson-to-bngpoint-function",
            functionid="d9a01773-6092-4cad-b331-ae725ae8fa88",
        )


    def remove_functions(apps, schema_editor):
        Function = apps.get_model("models", "Function")

        for fn in Function.objects.filter(pk__in=[
            "0434df8d-b98a-4b41-9a0a-68cd9214ad73",
            "96efa95a-1e2c-4562-ac1f-b415796f9f75"
            "39d627ae-6973-4ddb-8b62-1f0230e1e3f9",
            "d9a01773-6092-4cad-b331-ae725ae8fa88",
        ]):
            fn.delete()


    def add_widgets(apps, schema_editor):
        Widget = apps.get_model("models", "Widget")

        Widget.objects.update_or_create(
            widgetid="bcae8e90-09f7-4ae3-906b-7c7bb71a6ddf",
            name="bngpoint",
            component="views/components/widgets/bngpoint",
            defaultconfig={
                "placeholder": "Enter the centre point map reference of the resource."
            },
            helptext=None,
            datatype="bngcentrepoint"
        )

        Widget.objects.update_or_create(
            widgetid="31bc729d-6126-4301-8ec1-d6c4d98c68f8",
            name="photo-widget",
            datatype="file-list",
            component="views/components/widgets/photo",
            helptext=None,
            defaultconfig={
                "maxFilesize": "200",
                "acceptedFiles": "",
                "rerender": True
            }
        )
    
    def remove_widgets(apps, schema_editor):
        Widget = apps.get_model("models", "Widget")

        for widget in Widget.objects.filter(pk__in=[
            "bcae8e90-09f7-4ae3-906b-7c7bb71a6ddf",
            "31bc729d-6126-4301-8ec1-d6c4d98c68f8",
        ]):
            widget.delete()


    def add_datatypes(apps, schema_editor):
        Datatype = apps.get_model("models", "DDataType")
        Widget = apps.get_model("models", "Widget")

        bngpoint=Widget.objects.get(pk="bcae8e90-09f7-4ae3-906b-7c7bb71a6ddf")

        Datatype.objects.update_or_create(
            datatype="bngcentrepoint",
            iconclass="fa fa-location-arrow",
            modulename="bngcentrepoint.py",
            classname="BNGCentreDataType",
            defaultwidget=bngpoint,
            defaultconfig=None,
            configcomponent="views/components/datatypes/bngcentrepoint",
            configname="bngcentrepoint-datatype-config",
            isgeometric=False,
            issearchable=True,
        )

    def remove_datatypes(apps, schema_editor):
        Datatype = apps.get_model("models", "DDataType")

        for datatype in Datatype.objects.filter(datatype__in=[
            "bngcentrepoint",
        ]):
            datatype.delete()


    def add_search_components(apps, schema_editor):
        SearchComponent = apps.get_model("models", "SearchComponent")

        SearchComponent.objects.update_or_create(
            searchcomponentid="25ca3536-9eb4-4fd5-b2a5-badfd9a266de",
            name="BNG Filter",
            icon="fa fa-compass",
            modulename="bng-filter.py",
            classname="BngFilter",
            type="popup",
            componentpath="views/components/search/bng-filter",
            componentname="bng-filter",
            sortorder="0",
            enabled=True,
        )
    
    def remove_search_components(apps, schema_editor):
        SearchComponent = apps.get_model("models", "SearchComponent")

        for search_component in SearchComponent.objects.filter(pk__in=[
            "25ca3536-9eb4-4fd5-b2a5-badfd9a266de",
        ]):
            search_component.delete()


    add_map_source = """
INSERT INTO map_sources(name, source)
    VALUES ('select-application-area', '{
        "data": "/geojson?nodeid=1909956f-3a3b-11eb-ae99-f875a44e0e11", 
        "type": "vector", 
        "tiles": ["/application-areas/{z}/{x}/{y}.pbf"], 
        "minzoom": 6
    }');
        """

    remove_map_source = """
        DELETE FROM map_sources WHERE name = 'select-application-area';
        """

    operations = [
        migrations.RunPython(add_functions, remove_functions),
        migrations.RunPython(add_widgets, remove_widgets),
        migrations.RunPython(add_datatypes, remove_datatypes),
        migrations.RunPython(add_search_components, remove_search_components),
        migrations.RunPython(add_plugins, remove_plugins),
        migrations.RunPython(add_reports, remove_reports),
        migrations.RunSQL(add_map_source, remove_map_source),
    ]



from __future__ import absolute_import, unicode_literals
import os
from celery import Celery

import platform

if platform.system().lower() == "windows":
	os.environ.setdefault("FORKED_BY_MULTIPROCESSING","1")

os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'arches_her.settings')
app = Celery('arches_her')
app.config_from_object('django.conf:settings', namespace='CELERY')
app.autodiscover_tasks()



from django.views.generic import View
from django.db import connection
from django.http import HttpResponse, Http404
from arches.app.models import models
from arches.app.utils.permission_backend import get_restricted_instances
from arches.app.search.search_engine_factory import SearchEngineFactory


class ApplicationAreas(View):
    def get(self, request, zoom, x, y):
        nodeid = '1909956f-3a3b-11eb-ae99-f875a44e0e11' 
        if hasattr(request.user, "userprofile") is not True:
            models.UserProfile.objects.create(user=request.user)
        viewable_nodegroups = request.user.userprofile.viewable_nodegroups
        try:
            node = models.Node.objects.get(nodeid=nodeid, nodegroup_id__in=viewable_nodegroups)
            se = SearchEngineFactory().create()
            restricted_resource_ids = get_restricted_instances(request.user, search_engine=se)
            if len(restricted_resource_ids) == 0:
                restricted_resource_ids.append("10000000-0000-0000-0000-000000000001")  # This must have a uuid that will never be a resource id.
            restricted_resource_ids = tuple(restricted_resource_ids)
            
            with connection.cursor() as cursor:
                result = cursor.execute(
                    """SELECT ST_AsMVT(tile, 'app-area', 4096, 'geom', 'id') FROM (SELECT tileid,
                        id,
                        resourceinstanceid,
                        nodeid,
                        ST_AsMVTGeom(
                            geom,
                            TileBBox(%s, %s, %s, 3857)
                        ) AS geom,
                        1 AS total
                    FROM geojson_geometries
                    WHERE nodeid = %s and resourceinstanceid not in %s and (geom && ST_TileEnvelope(%s, %s, %s))) AS tile;""",
                    [zoom, x, y, nodeid, restricted_resource_ids, zoom, x, y],
                )
                result = bytes(cursor.fetchone()[0]) if result is None else result
            return HttpResponse(result, content_type="application/x-protobuf")
    
        except models.Node.DoesNotExist:
            return HttpResponse(status=503)


from uuid import UUID
from arches.app.functions.base import BaseFunction
from arches.app.models import models
from arches.app.models.tile import Tile
from arches.app.models.system_settings import settings
from django.db.models.functions import Cast
from django.db.models import Max, IntegerField
from django.db import connection
import logging
import json
from datetime import datetime

details = {
    "name": "Generate Unique References",
    "type": "node",
    "description": "Checks for Simple UID and Resource ID nodes populated and, if not populated, generates them",
    "defaultconfig": {
        "simpleuid_node": "",
        "resourceid_node": "",
        "uniqueresource_nodegroup": "",
        "triggering_nodegroups": [],
        "nodegroup_nodes": [],
    },
    "classname": "GenerateUniqueReferences",
    "component": "views/components/functions/generate-unique-references-function",
    "functionid": "39d627ae-6973-4ddb-8b62-1f0230e1e3f9",
}


class GenerateUniqueReferences(BaseFunction):
    # def __init__(self,):
    # super(GenerateUniqueReferences, self).__init__()
    # self.logger = logging.getLogger(__name__)

    def get(self):
        raise NotImplementedError

    def save(self, tile, request, context=None):
        self.logger = logging.getLogger(__name__)
        try:

            def get_next_simple_id():
                """
                Gets the max id for all simple_id nodes across all graphs where the GenerateUniqueReferences function has been configured
                """
                self.logger.debug(
                    str(datetime.now())
                    + " DEBUG: get_next_simple_id CALLED ========================================"
                )

                if simpleid_nextval_table_exists():
                    return get_next_simpleid()

                simpleid_node_info = get_simple_id_nodeinfo(details["functionid"])

                # Don't like doing direct SQL but the django model is too slow.
                # The input values are pulled from the database using the function configs.
                sql_node_str = ""
                sql_nodegroups = []
                sql_params = []
                for nodeinfo in simpleid_node_info:
                    sid = nodeinfo["simpleid"]
                    ngid = nodeinfo["unique_ng_id"]

                    # UUID parse to give a sql injection protection
                    try:
                        x = UUID(sid)
                        x = UUID(ngid)
                    except Exception:
                        raise TypeError(
                            "Expected UUID values are not parsing correctly"
                        )

                    sql_node_str = sql_node_str + "t.tiledata ->> %s::text,"
                    sql_params.append(str(sid))
                    sql_nodegroups.append(ngid)

                sql = f"""
                SELECT hobs.simple_id::int
                    FROM (
                        SELECT
                        COALESCE(
                            {sql_node_str}
                        '0') AS simple_id
                        FROM tiles t
                        WHERE nodegroupid IN %s
                    ) hobs
                    ORDER BY hobs.simple_id::int DESC
                    LIMIT 1
                """

                sql_nodegroups.append("11111111-1111-1111-1111-111111111111")
                sql_params.append(tuple(sql_nodegroups))

                with connection.cursor() as cursor:
                    cursor.execute(sql, sql_params)
                    ret = cursor.fetchone()

                try:
                    ret = int(ret[0]) + 1
                except:
                    ret = 1

                if not simpleid_nextval_table_exists():
                    create_simpleid_nextval_table(start=ret + 1)

                self.logger.debug(
                    str(datetime.now())
                    + " DEBUG: get_next_simple_id COMPLETE ({0})========================================".format(
                        ret
                    )
                )
                return ret

            def create_simpleid_nextval_table(start=1):
                cursor = connection.cursor()
                sql_params = [start]
                # New code to create table and sequence
                create_table_query = """
                    CREATE SEQUENCE simpleid_nextval_id_seq MINVALUE 1 START %s;
                    
                    CREATE TABLE simpleid_nextval
                    (
                        ID integer NOT NULL DEFAULT nextval('simpleid_nextval_id_seq'),
                        CONSTRAINT simpleid_nextval_pkey PRIMARY KEY (ID)
                    );
                    
                    ALTER SEQUENCE simpleid_nextval_id_seq OWNED BY simpleid_nextval.ID;
                """

                cursor.execute(create_table_query, sql_params)

            def simpleid_nextval_table_exists():
                with connection.cursor() as cursor:
                    cursor.execute(
                        """
                        SELECT EXISTS (
                            SELECT FROM information_schema.tables 
                            WHERE  table_schema = 'public'
                            AND    table_name   = 'simpleid_nextval'
                        );
                    """
                    )
                    return bool(cursor.fetchone()[0])

            def get_next_simpleid():
                with connection.cursor() as cursor:
                    cursor.execute(
                        """
                        TRUNCATE TABLE simpleid_nextval;
                        INSERT INTO simpleid_nextval DEFAULT VALUES RETURNING id;
                    """
                    )
                    return int(cursor.fetchone()[0])

            def get_simple_id_nodeinfo(functionid=None):
                """
                This will fetch all the set simpleid nodes configured across the graphs
                """
                if functionid is None:
                    raise Exception("Functionid is required")

                nodeinfos = []

                funcs = models.FunctionXGraph.objects.filter(function_id=functionid)
                for fn in funcs:
                    try:
                        nodeinfos.append(
                            {
                                "simpleid": fn.config["simpleuid_node"],
                                "unique_ng_id": fn.config["uniqueresource_nodegroup"],
                            }
                        )
                    except:
                        pass

                return nodeinfos

            resourceIdValue = tile.resourceinstance_id
            simpleNode = self.config["simpleuid_node"]
            resourceIdNode = self.config["resourceid_node"]
            refNodegroup = self.config["uniqueresource_nodegroup"]

            def check_and_populate_uids(
                currentTile, simpleid_node, resid_node, resourceidval
            ):
                """
                Checks the input tile to see if it contains the correct resource id and simpleid
                values.  If not, populates these ids.
                """

                def populate_simple_id(currentTile, simple_node_id):
                    nextsimpleval = get_next_simple_id()            
                    currentTile.data[simple_node_id] = nextsimpleval
                    
                def format_string_value(id_string_value):
                    languages = models.Language.objects.all()
                    default_language = languages.get(code=settings.LANGUAGE_CODE)
                    return {default_language.code: {"value": id_string_value, "direction": default_language.default_direction}}

                try:
                    if currentTile.data[simpleid_node] is not None:
                        if currentTile.data[simpleid_node] != 0:
                            try:
                                x = int(currentTile.data[simpleid_node])
                                self.logger.debug(
                                    "Resource "
                                    + str(resourceidval)
                                    + "has valid simpleid: "
                                    + str(x)
                                )
                                pass
                            except:
                                populate_simple_id(currentTile, simpleid_node)
                        else:
                            populate_simple_id(currentTile, simpleid_node)
                    else:
                        populate_simple_id(currentTile, simpleid_node)

                    if currentTile.data[resid_node] is not None:
                        try:
                            UUID(currentTile.data[resid_node])
                            pass
                        except:
                            currentTile.data[resid_node] = format_string_value(str(resourceidval))
                    else:
                        currentTile.data[resid_node] = format_string_value(str(resourceidval))

                    return True

                except (
                    KeyboardInterrupt,
                    SystemExit,
                    ImportError,
                    RuntimeError,
                    SyntaxError,
                ) as c:
                    self.logger.critical(str(c))
                    return False

                except (
                    AttributeError,
                    EOFError,
                    LookupError,
                    NameError,
                    MemoryError,
                    ValueError,
                    IOError,
                ) as e:
                    self.logger.error(str(e))
                    return False

                except Warning as w:
                    self.logger.warning(str(w))
                    return False

                except Exception as ex:
                    self.logger.error(str(ex))
                    return False

            # if the current tile context is a refNG (i.e. it has trigger its own save then don't trigger another save)
            if str(tile.nodegroup_id) == refNodegroup:
                check_and_populate_uids(
                    tile, simpleNode, resourceIdNode, resourceIdValue
                )
                return

            previously_saved_tiles = Tile.objects.filter(
                nodegroup_id=refNodegroup, resourceinstance_id=resourceIdValue
            )

            if len(previously_saved_tiles) > 0:
                for p in previously_saved_tiles:
                    try:
                        if (
                            check_and_populate_uids(
                                p, simpleNode, resourceIdNode, resourceIdValue
                            )
                            == True
                        ):
                            p.save()
                        else:
                            self.logger.debug(
                                "Error.  Could not save Unique Identifiers tile."
                            )
                    except (
                        KeyboardInterrupt,
                        SystemExit,
                        ImportError,
                        RuntimeError,
                        SyntaxError,
                    ) as c:
                        self.logger.critical(str(c))

                    except (
                        AttributeError,
                        EOFError,
                        LookupError,
                        NameError,
                        MemoryError,
                        ValueError,
                        IOError,
                    ) as e:
                        self.logger.error(str(e))

                    except Warning as w:
                        self.logger.warning(str(w))

                    except Exception as ex:
                        self.logger.error(str(ex))
            else:
                newRefTile = Tile().get_blank_tile_from_nodegroup_id(
                    refNodegroup, resourceid=resourceIdValue, parenttile=None
                )
                if (
                    check_and_populate_uids(
                        newRefTile, simpleNode, resourceIdNode, resourceIdValue
                    )
                    == True
                ):
                    newRefTile.save()
                else:
                    self.logger.debug("Error.  Could not save Unique Identifiers tile.")

            return

        except (
            KeyboardInterrupt,
            SystemExit,
            ImportError,
            RuntimeError,
            SyntaxError,
        ) as c:
            self.logger.critical(str(c))

        except (
            AttributeError,
            EOFError,
            LookupError,
            NameError,
            MemoryError,
            ValueError,
            IOError,
        ) as e:
            self.logger.error(str(e))

        except Warning as w:
            self.logger.warning(str(w))

        except Exception as ex:
            self.logger.error(str(ex))

    def delete(self, tile, request):
        raise NotImplementedError

    def on_import(self, tile):
        raise NotImplementedError

    def after_function_save(self, tile, request):
        raise NotImplementedError


ELASTICSEARCH_HOSTS = [{"scheme": "http", "host": "localhost", "port": 9200}]
EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'  #<-- Only need to uncomment this for testing without an actual email server
SEARCH_EXPORT_IMMEDIATE_DOWNLOAD_THRESHOLD = 2000
ACCESSIBILITY_MODE = False


from django.urls import include, path, re_path
from django.conf import settings
from django.conf.urls.static import static
#from django.conf.urls.i18n import i18n_patterns
from arches.app.views.plugin import PluginView
from arches_her.views.file_template import FileTemplateView
from arches_her.views.active_consultations import ActiveConsultationsView
from arches_her.views.index import IndexView
from arches_her.views.map import ApplicationAreas

uuid_regex = settings.UUID_REGEX

urlpatterns = [
    re_path(r'^$', IndexView.as_view(), name='root'),
    re_path(r'^index.htm', IndexView.as_view(), name='home'),
    path('', include('arches.urls')),
    re_path(r'^filetemplate', FileTemplateView.as_view(), name='filetemplate'),
    re_path(r'^plugins/active-consultations$', PluginView.as_view(), name='active-consultations'),
    re_path(r'^activeconsultations', ActiveConsultationsView.as_view(), name='activeconsultations'),
    re_path(r'^plugins/application-area', PluginView.as_view(), name='application-area'),
    re_path(r'^plugins/consultation-workflow', PluginView.as_view(), name='consultation-workflow'),
    re_path(r'^plugins/site-visit', PluginView.as_view(), name='site-visit'),
    re_path(r'^plugins/correspondence-workflow', PluginView.as_view(), name='correspondence-workflow'),
    re_path(r'^plugins/communication-workflow', PluginView.as_view(), name='communication-workflow'),
    re_path(r'^plugins/init-workflow', PluginView.as_view(), name='init-workflow'),
    re_path(r"^application-areas/(?P<zoom>[0-9]+|\{z\})/(?P<x>[0-9]+|\{x\})/(?P<y>[0-9]+|\{y\}).pbf$", ApplicationAreas.as_view(), name="application-areas"),
] + static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)

# if settings.SHOW_LANGUAGE_SWITCH is True:
#     urlpatterns = i18n_patterns(*urlpatterns)


from arches.app.datatypes.base import BaseDataType
from arches.app.models.models import Widget
from arches.app.models.system_settings import settings
from arches.app.search.elasticsearch_dsl_builder import Match, Exists
from arches.app.search.search_term import SearchTerm
import logging

bngpoint = Widget.objects.get(name="bngpoint")

details = {
    "datatype": "bngcentrepoint",
    "iconclass": "fa fa-location-arrow",
    "modulename": "bngcentrepoint.py",
    "classname": "BNGCentreDataType",
    "defaultwidget": bngpoint,
    "defaultconfig": None,
    "configcomponent": "views/components/datatypes/bngcentrepoint",
    "configname": "bngcentrepoint-datatype-config",
    "isgeometric": False,
    "issearchable": True,
}


class BNGCentreDataType(BaseDataType):
    def validate(self, value, row_number=None, source=None, node=None, nodeid=None, strict=False, **kwargs):

        errors = []
        gridSquareArray = [
            "NA",
            "NB",
            "NC",
            "ND",
            "NE",
            "NF",
            "NG",
            "NH",
            "NJ",
            "NK",
            "NL",
            "NM",
            "NN",
            "NO",
            "NP",
            "NR",
            "NS",
            "NT",
            "NU",
            "NV",
            "NW",
            "NX",
            "NY",
            "NZ",
            "OA",
            "OB",
            "OF",
            "OG",
            "OL",
            "OM",
            "OQ",
            "OR",
            "OV",
            "OW",
            "SA",
            "SB",
            "SC",
            "SD",
            "SE",
            "SF",
            "SG",
            "SH",
            "SJ",
            "SK",
            "SL",
            "SM",
            "SN",
            "SO",
            "SP",
            "SQ",
            "SR",
            "SS",
            "ST",
            "SU",
            "SV",
            "SW",
            "SX",
            "SY",
            "SZ",
            "TA",
            "TB",
            "TF",
            "TG",
            "TL",
            "TM",
            "TQ",
            "TR",
            "TV",
            "TW",
        ]
        if value is not None:
            try:
                # CS - Validation for datatype.  Replicates functionality in widget which will be removed once datatype validation is fixed.
                firstTwoCharacters = value[0:2]
                numberElement = value[2:]

                firstTwoCharacters in gridSquareArray
                isinstance(int(numberElement), int)
                len(value) == 12
            except Exception:
                errors.append({"type": "ERROR", "message": "Issue with input data"})

        return errors

    def append_to_document(self, document, nodevalue, nodeid, tile, provisional=False):

        document["strings"].append({"string": nodevalue, "nodegroup_id": tile.nodegroup_id})

    def get_search_terms(self, nodevalue, nodeid=None):
        return [SearchTerm(value=nodevalue)]

    def append_search_filters(self, value, node, query, request):
        self.logger = logging.getLogger(__name__)
        try:
            if value["op"] == "null" or value["op"] == "not_null":
                self.append_null_search_filters(value, node, query, request)
            elif value["val"] != "":
                match_type = "phrase_prefix" if "~" in value["op"] else "phrase"
                match_query = Match(field="tiles.data.%s" % (str(node.pk)), query=value["val"], type=match_type)
                if "!" in value["op"]:
                    query.must_not(match_query)
                    query.filter(Exists(field="tiles.data.%s" % (str(node.pk))))
                else:
                    query.must(match_query)
        except KeyError as e:
            self.logger.error(str(e))




import uuid
from arches.app.functions.base import BaseFunction
from arches.app.models.system_settings import settings
from arches.app.models import models
from arches.app.models.tile import Tile
from arches.app.models.resource import Resource
from django.contrib.gis.geos import GEOSGeometry
from django.db import connection, transaction
import json
from datetime import datetime


details = {
    "name": "BNG Point to GeoJSON",
    "type": "node",
    "description": "Pushes the geometry from a BNG Point node to a related GeoJSON node",
    "defaultconfig": {"bng_node": "", "geojson_node": "", "bng_nodegroup": "", "geojson_nodegroup": "", "triggering_nodegroups": []},
    "classname": "BNGPointToGeoJSON",
    "component": "views/components/functions/bngpoint-to-geojson-function",
    "functionid": "0434df8d-b98a-4b41-9a0a-68cd9214ad73",
}


class BNGPointToGeoJSON(BaseFunction):
    def get(self):
        raise NotImplementedError

    def save_geojson(self, tile, request, is_function_save_method=True):
        """Finds the equivalen GeoJSON for a BNG Alphanumeric value and saves that value to the
        geojson nodegroup of the tile.

        Args:
            self : BNGPointToGeoJSON object.

            tile : Tile to attach / amend geojson_nodegroup of.

            request : WSGI Request used to varify call is result of user action. N.B. Function Returns if empty.

            is_function_save_method : a bool stating whether the function calling it is the save function.
        """

        # First let's check if this call is as a result of an inbound request (user action) or
        # as a result of the complementary GeoJSONToBNGPoint function saving a new BngPoint.
        if request is None and is_function_save_method == True:
            return

        bngValueReturned = ""
        gridSquare = {
            "NA": [0, 9],
            "NB": [1, 9],
            "NC": [2, 9],
            "ND": [3, 9],
            "NE": [4, 9],
            "OA": [5, 9],
            "OB": [6, 9],
            "NF": [0, 8],
            "NG": [1, 8],
            "NH": [2, 8],
            "NJ": [3, 8],
            "NK": [4, 8],
            "OF": [5, 8],
            "OG": [6, 8],
            "NL": [0, 7],
            "NM": [1, 7],
            "NN": [2, 7],
            "NO": [3, 7],
            "NP": [4, 7],
            "OL": [5, 7],
            "OM": [6, 7],
            "NQ": [0, 6],
            "NR": [1, 6],
            "NS": [2, 6],
            "NT": [3, 6],
            "NU": [4, 6],
            "OQ": [5, 6],
            "OR": [6, 6],
            "NV": [0, 5],
            "NW": [1, 5],
            "NX": [2, 5],
            "NY": [3, 5],
            "NZ": [4, 5],
            "OV": [5, 5],
            "OW": [6, 5],
            "SA": [0, 4],
            "SB": [1, 4],
            "SC": [2, 4],
            "SD": [3, 4],
            "SE": [4, 4],
            "TA": [5, 4],
            "TB": [6, 4],
            "SF": [0, 3],
            "SG": [1, 3],
            "SH": [2, 3],
            "SJ": [3, 3],
            "SK": [4, 3],
            "TF": [5, 3],
            "TG": [6, 3],
            "SL": [0, 2],
            "SM": [1, 2],
            "SN": [2, 2],
            "SO": [3, 2],
            "SP": [4, 2],
            "TL": [5, 2],
            "TM": [6, 2],
            "SQ": [0, 1],
            "SR": [1, 1],
            "SS": [2, 1],
            "ST": [3, 1],
            "SU": [4, 1],
            "TQ": [5, 1],
            "TR": [6, 1],
            "SV": [0, 0],
            "SW": [1, 0],
            "SX": [2, 0],
            "SY": [3, 0],
            "SZ": [4, 0],
            "TV": [5, 0],
            "TW": [6, 0],
        }

        bngnode = self.config["bng_node"]
        geojsonNode = self.config["geojson_node"]
        bngValueReturned = tile.data[bngnode]

        if bngValueReturned != None:
            """
            The following section turns the alphanumberic BNG value in the tile into a point geometry object and then transforms that object
            into WGS 1984 long/lat projection system.
            """

            dt = datetime.now()
            gridSquareLetters = bngValueReturned[0:2]
            bngValueNumbers = bngValueReturned[2:]
            splitSection = int(len(bngValueNumbers) / 2)
            gridSquareNumbers = gridSquare[gridSquareLetters]
            eastingValue = str(gridSquareNumbers[0]) + str(bngValueNumbers[:splitSection])
            northingValue = str(gridSquareNumbers[1]) + str(bngValueNumbers[splitSection:])
            osgb36PointString = "POINT (" + eastingValue + " " + northingValue + ")"
            osgb36Point = GEOSGeometry(osgb36PointString, srid=27700)
            osgb36Point.transform(4326, False)
            pointGeoJSON = json.loads(osgb36Point.geojson)

            """
                This section creates a geojson object required in the format required by Arches.  The date and time the object was
                created has also been added in the feature's properties.
            """

            uuidForRecord = uuid.uuid4().hex
            bngFeature = {
                "geometry": pointGeoJSON,
                "type": "Feature",
                "id": str(uuidForRecord),
                "properties": {"datetime": dt.strftime("%d/%m/%Y %H:%M:%S"), "bngref": str(bngValueReturned)},
            }

            geometryValue = {"type": "FeatureCollection", "features": [bngFeature]}

            geometryValueJson = geometryValue

            """
            The Tile.objects.filter function from tiles.py is called to return any tiles with the geojson_nodegroup value
            as the nodegroup_id and the current tile's resource instance ID as its resourceinstance_id value; any tiles returned
            are added to the previously_saved_tiles variable.

            If there are tiles returned then the new geojson object overwrites the current value.

            If there are no tiles returned, a new tile is created for the geojson nodegroup using tile.py's get_blank_tile
            function.  If there is a key within the data object in the new node with the same id as the geojson_nodegroup value
            then that key/value pair are deleted.  The geojson object is set at the value to the key which has the value of the geojson_node
            value.

            The new tile is saved and then the mv_geojson_geoms materialised view is refreshed so the point geometry will be displayed
            on the Search map.
            """
            if self.config["geojson_nodegroup"] == str(tile.nodegroup_id):
                tile.data[geojsonNode] = geometryValueJson
            else:
                previously_saved_tiles = Tile.objects.filter(
                    nodegroup_id=self.config["geojson_nodegroup"], resourceinstance_id=tile.resourceinstance_id
                )

                if len(previously_saved_tiles) > 0:
                    for p in previously_saved_tiles:
                        old_geojson = p.data[geojsonNode]
                        p.data[geojsonNode] = geometryValueJson

                        for f in old_geojson["features"]:
                            if "bngref" not in f["properties"]:
                                p.data[geojsonNode]["features"].append(f)

                        p.save()
                else:
                    new_geojson_tile = Tile().get_blank_tile_from_nodegroup_id(
                        self.config["geojson_nodegroup"], resourceid=tile.resourceinstance_id, parenttile=tile.parenttile
                    )
                    new_geojson_tile.data[self.config["geojson_node"]] = geometryValueJson

                    if self.config["geojson_nodegroup"] in new_geojson_tile.data:
                        del new_geojson_tile.data[self.config["geojson_nodegroup"]]

                    new_geojson_tile.save()

            cursor = connection.cursor()
            sql = """
                    SELECT * FROM refresh_geojson_geometries();
                """
            cursor.execute(sql)  #

        else:
            pass

        return

    def save(self, tile, request, context=None):

        self.save_geojson(tile=tile, request=request, is_function_save_method=True)
        return

    def post_save(self, *args, **kwargs):
        raise NotImplementedError

    def delete(self, tile, request):
        raise NotImplementedError

    def on_import(self, tile):
        self.save_geojson(tile=tile, request=None, is_function_save_method=False)
        return

    def after_function_save(self, tile, request):
        raise NotImplementedError



import logging
import json
from operator import invert
from django.contrib.gis.geos import GEOSGeometry
from django.db import connection
from django.utils.translation import gettext as _
from uuid import uuid4
from arches.app.models.system_settings import settings
from arches.app.utils.betterJSONSerializer import JSONSerializer, JSONDeserializer
from arches.app.search.elasticsearch_dsl_builder import Bool, Nested, Terms, GeoShape
from arches.app.search.components.base import BaseSearchFilter

logger = logging.getLogger(__name__)

details = {
    "searchcomponentid": "",
    "name": "BNG Filter",
    "icon": "fa fa-compass",
    "modulename": "bng-filter.py",
    "classname": "BngFilter",
    "type": "popup",
    "componentpath": "views/components/search/bng-filter",
    "componentname": "bng-filter",
    "sortorder": "0",
    "enabled": True,
}


class BngFilter(BaseSearchFilter):
    def append_dsl(self, search_results_object, permitted_nodegroups, include_provisional):

        search_query = Bool()
        querysting_params = self.request.GET.get(details["componentname"], "")
        bng_filter = JSONDeserializer().deserialize(querysting_params)
        bng = bng_filter["bng"].upper()
        buffer = bng_filter["buffer"]
        inverted = bng_filter["inverted"]

        # bng should have an even number of chars to be a correct grid
        if len(bng) % 2 == 0:
            spatial_filter = self.build_geojson_from_bng(bng, buffer)
            print(json.dumps(spatial_filter))
            if "features" in spatial_filter:
                if len(spatial_filter["features"]) > 0:
                    grid_square_geom = spatial_filter["features"][0]["geometry"]
                    try:
                        grid_square_buffer = spatial_filter["features"][1]["geometry"]
                    except:
                        grid_square_buffer = None

                    dsl_geom = grid_square_geom if grid_square_buffer is None else grid_square_buffer

                    geoshape = GeoShape(
                        field="geometries.geom.features.geometry", type=dsl_geom["type"], coordinates=dsl_geom["coordinates"]
                    )

                    spatial_query = Bool()
                    if inverted is True:
                        spatial_query.must_not(geoshape)
                    else:
                        spatial_query.filter(geoshape)

                    # get the nodegroup_ids that the user has permission to search
                    spatial_query.filter(Terms(field="geometries.nodegroup_id", terms=permitted_nodegroups))

                    if include_provisional is False:
                        spatial_query.filter(Terms(field="geometries.provisional", terms=["false"]))

                    elif include_provisional == "only provisional":
                        spatial_query.filter(Terms(field="geometries.provisional", terms=["true"]))

                    search_query.filter(Nested(path="geometries", query=spatial_query))
        else:
            logger.warn(_(f"BNG Filter: BNG is not valid - must be an even number of chars ({bng})"))

        search_results_object["query"].add_query(search_query)

        if details["componentname"] not in search_results_object:
            search_results_object[details["componentname"]] = {}

        try:
            search_results_object[details["componentname"]]["grid_square"] = spatial_filter
        except NameError:
            logger.info(_("Feature geometry is not defined"))

    def view_data(self):
        """
        data that the view should gather to pass to the front end

        """

        pass

    def post_search_hook(self, search_results_object, results, permitted_nodegroups):
        """
        code to run after the search results have been retrieved

        """

        pass

    def build_geojson_from_bng(self, bng_value, buffer_value=0):

        geometryValue = {"type": "FeatureCollection", "features": []}

        if bng_value != None:

            grid_square = self.bng_grid_square()

            gridSquareLetters = bng_value[0:2]
            bngValueNumbers = bng_value[2:]
            splitSection = int(len(bngValueNumbers) / 2)
            gridSquareNumbers = grid_square[gridSquareLetters]
            eastingValue = f"{str(gridSquareNumbers[0])}{str(bngValueNumbers[:splitSection])}"
            northingValue = f"{str(gridSquareNumbers[1])}{str(bngValueNumbers[splitSection:])}"

            xmin = self.pad_coord(eastingValue, "0", 6)
            xmax = self.pad_coord(eastingValue, "9", 6)
            ymin = self.pad_coord(northingValue, "0", 6)
            ymax = self.pad_coord(northingValue, "9", 6)
            isPolygon = True
            if xmin == xmax and ymin == ymax:
                isPolygon = False
                wkt_geom = "POINT(" + xmin + " " + ymax + ")"
            else:
                wkt_geom = (
                    "POLYGON(("
                    + xmin
                    + " "
                    + ymin
                    + ","
                    + xmin
                    + " "
                    + ymax
                    + ","
                    + xmax
                    + " "
                    + ymax
                    + ","
                    + xmax
                    + " "
                    + ymin
                    + ","
                    + xmin
                    + " "
                    + ymin
                    + "))"
                )

            bng_geom = self.transform_to_wgs84(GEOSGeometry(wkt_geom, srid=27700), from_srid=27700)
            bng_geojson = json.loads(bng_geom.geojson)

            uuidForRecord = uuid4()
            bng_feature = {
                "geometry": bng_geojson,
                "type": "Feature",
                "id": str(uuidForRecord),
                "properties": {"type": "grid_square", "bngref": str(bng_value)},
            }

            geometryValue["features"].append(bng_feature)

            if buffer_value > 0:
                bng_buffer_polygon = _buffer(bng_geom, buffer_value)
                bng_buffer_geojson = json.loads(bng_buffer_polygon.geojson)
                uuidForRecord = uuid4()
                bng_buffer_feature = {
                    "geometry": bng_buffer_geojson,
                    "type": "Feature",
                    "id": str(uuidForRecord),
                    "properties": {"type": "grid_square_buffer", "bngref": str(bng_value), "buffer": buffer_value},
                }
                geometryValue["features"].append(bng_buffer_feature)

        return geometryValue

    def pad_coord(self, coord, pad_value, pad_length):
        """
        This function right pads a string with a given value to a given length.
        """
        if len(coord) < pad_length:
            coord = coord + pad_value * (pad_length - len(coord)) + f".{pad_value}"
        return coord

    def transform_to_wgs84(self, geom, from_srid):
        to_srid = 4326
        with connection.cursor() as cursor:
            # Transform geom to the analysis SRID, buffer it, and transform it back to wgs84
            cursor.execute(
                """SELECT ST_TRANSFORM(ST_SETSRID(%s::geometry, %s), %s)""",
                (geom.hex.decode("utf-8"), from_srid, to_srid),
            )
            res = cursor.fetchone()
            geom = GEOSGeometry(res[0], srid=4326)
        return geom

    def bng_grid_square(self):
        return {
            "HO": [3, 12],
            "HP": [4, 12],
            "HT": [3, 11],
            "HU": [4, 11],
            "HW": [1, 10],
            "HX": [2, 10],
            "HY": [3, 10],
            "HZ": [4, 10],
            "NA": [0, 9],
            "NB": [1, 9],
            "NC": [2, 9],
            "ND": [3, 9],
            "NE": [4, 9],
            "OA": [5, 9],
            "OB": [6, 9],
            "NF": [0, 8],
            "NG": [1, 8],
            "NH": [2, 8],
            "NJ": [3, 8],
            "NK": [4, 8],
            "OF": [5, 8],
            "OG": [6, 8],
            "NL": [0, 7],
            "NM": [1, 7],
            "NN": [2, 7],
            "NO": [3, 7],
            "NP": [4, 7],
            "OL": [5, 7],
            "OM": [6, 7],
            "NQ": [0, 6],
            "NR": [1, 6],
            "NS": [2, 6],
            "NT": [3, 6],
            "NU": [4, 6],
            "OQ": [5, 6],
            "OR": [6, 6],
            "NV": [0, 5],
            "NW": [1, 5],
            "NX": [2, 5],
            "NY": [3, 5],
            "NZ": [4, 5],
            "OV": [5, 5],
            "OW": [6, 5],
            "SA": [0, 4],
            "SB": [1, 4],
            "SC": [2, 4],
            "SD": [3, 4],
            "SE": [4, 4],
            "TA": [5, 4],
            "TB": [6, 4],
            "SF": [0, 3],
            "SG": [1, 3],
            "SH": [2, 3],
            "SJ": [3, 3],
            "SK": [4, 3],
            "TF": [5, 3],
            "TG": [6, 3],
            "SL": [0, 2],
            "SM": [1, 2],
            "SN": [2, 2],
            "SO": [3, 2],
            "SP": [4, 2],
            "TL": [5, 2],
            "TM": [6, 2],
            "SQ": [0, 1],
            "SR": [1, 1],
            "SS": [2, 1],
            "ST": [3, 1],
            "SU": [4, 1],
            "TQ": [5, 1],
            "TR": [6, 1],
            "SV": [0, 0],
            "SW": [1, 0],
            "SX": [2, 0],
            "SY": [3, 0],
            "SZ": [4, 0],
            "TV": [5, 0],
            "TW": [6, 0],
        }


def _buffer(geojson, width=0):
    geojson = JSONSerializer().serialize(geojson)
    geom = GEOSGeometry(geojson, srid=4326)

    try:
        width = float(width)
    except Exception:
        width = 0

    if width > 0:
        with connection.cursor() as cursor:
            # Transform geom to the analysis SRID, buffer it, and transform it back to wgs84
            cursor.execute(
                """SELECT ST_TRANSFORM(
                    ST_BUFFER(ST_TRANSFORM(ST_SETSRID(%s::geometry, 4326), %s), %s),
                4326)""",
                (geom.hex.decode("utf-8"), settings.ANALYSIS_COORDINATE_SYSTEM_SRID, width),
            )
            res = cursor.fetchone()
            geom = GEOSGeometry(res[0], srid=4326)
    return geom


"""
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

from django.http import HttpResponseNotFound
from django.views.generic import View
from django.core.paginator import Paginator
from arches.app.utils.response import JSONResponse
from arches.app.models import models
from arches.app.models.resource import Resource
from arches.app.models.tile import Tile
from arches.app.datatypes.datatypes import DataTypeFactory
import json
from datetime import datetime
from arches.app.models.system_settings import settings


class ActiveConsultationsView(View):
    def __init__(self):
        self.cons_details_nodegroupid = "771bb1e2-8895-11ea-8446-f875a44e0e11"  # "Consultation Type" nodegroup, needs to be updated but not used
        self.consultation_graphid = "8d41e49e-a250-11e9-9eab-00224800b26d"
        self.cons_status_bool_nodeid = "6a773228-db20-11e9-b6dd-784f435179ea"
        self.active_cons_node_list = {  # if this is not up-to-date sorting will break
            "Geospatial Coordinates": "b949053a-184f-11eb-ac4a-f875a44e0e11",
            "Consultation Name": "4ad69684-951f-11ea-b5c3-f875a44e0e11",
            "Consultation Type": "771bb1e2-8895-11ea-8446-f875a44e0e11",
            "Proposal Text": "1b0e15ec-8864-11ea-8493-f875a44e0e11",
            "Target Date Start": "7224417b-893a-11ea-b383-f875a44e0e11",
            "Casework Officer": "4ea4a197-184f-11eb-9152-f875a44e0e11",
            "Log Date": "40eff4cd-893a-11ea-b0cc-f875a44e0e11",
        }

    def get(self, request):
        page_num = 1 if request.GET.get("page") == "" else int(request.GET.get("page"))
        order_param = request.GET.get("order")
        keyword = (
            None
            if request.GET.get("keyword") == "" or request.GET.get("keyword") == None
            else (request.GET.get("keyword"))
        )
        datatype_factory = DataTypeFactory()
        consultation_status_tiles = Tile.objects.filter(
            nodegroup_id=self.cons_status_bool_nodeid
        )
        include_list = [
            tile.resourceinstance_id
            for tile in consultation_status_tiles
            if tile.data[self.cons_status_bool_nodeid] is not False
        ]
        filtered_consultations = Resource.objects.filter(
            graph_id=self.consultation_graphid, resourceinstanceid__in=include_list
        )

        order_config = {  # if this is not up-to-date sorting will break
            "Log Date: Newest to Oldest": ("Log Date", True),
            "Log Date: Oldest to Newest": ("Log Date", False),
            "Target Date: Newest to Oldest": ("Target Date Start", True),
            "Target Date: Oldest to Newest": ("Target Date Start", False),
            "Casework Officer: A to Z": ("Casework Officer", False),
            "Casework Officer: Z to A": ("Casework Officer", True),
            "Consultation Type: A to Z": ("Consultation Type", False),
            "Consultation Type: Z to A": ("Consultation Type", True),
            "Consultation Name: A to Z": ("Consultation Name", False),
            "Consultation Name: Z to A": ("Consultation Name", True),
        }

        search_results_setting_nodeid = "d0987de3-fad8-11e6-a434-6c4008b05c4c"
        search_results_setting_nodegroupid = "d0987880-fad8-11e6-8cce-6c4008b05c4c"
        page_ct_tile = Tile.objects.get(nodegroup_id=search_results_setting_nodegroupid)
        page_ct = page_ct_tile.data[search_results_setting_nodeid]

        if filtered_consultations is not None:
            if page_num == -1:
                grouped_tile_list = build_resource_dict(
                    filtered_consultations,
                    self.active_cons_node_list,
                    datatype_factory,
                    keyword=keyword,
                    layout="table",
                )
            elif page_num >= 1:
                grouped_tile_list = build_resource_dict(
                    filtered_consultations,
                    self.active_cons_node_list,
                    datatype_factory,
                    keyword=keyword,
                )
            if order_param in order_config:
                date_formats = settings.DATE_FORMATS["Python"]
                date_only_format = next(
                    (
                        format
                        for format in date_formats
                        if "%Y" in format
                        and "%m" in format
                        and "%d" in format
                        and "%H" not in format
                        and "%M" not in format
                        and "%S" not in format
                    ),
                    None,
                )
                try:
                    grouped_tile_list = sorted(
                        grouped_tile_list,
                        key=lambda resource: (
                            resource.get(order_config[order_param][0]) is None,
                            (
                                datetime.strptime(
                                    resource.get(order_config[order_param][0]),
                                    date_only_format,
                                ).strftime("%Y%m%d")
                                if "Date" in order_config[order_param][0]
                                and resource.get(order_config[order_param][0])
                                and resource.get(order_config[order_param][0]) != ""
                                else resource.get(order_config[order_param][0])
                            ),
                        ),
                        reverse=order_config[order_param][1],
                    )
                except KeyError as e:
                    print("Error: ", e)
            if page_num == -1:
                return JSONResponse({"results": grouped_tile_list})
            else:
                return self.get_paginated_data(grouped_tile_list, page_ct, page_num)

        return HttpResponseNotFound()

    def get_paginated_data(self, grouped_tile_list, page_ct, page_num):
        paginator = Paginator(grouped_tile_list, page_ct)
        page_results = paginator.page(page_num)
        if page_results.has_next() is True:
            next_page_number = page_results.next_page_number()
        else:
            next_page_number = False
        if page_results.has_previous() is True:
            prev_page_number = page_results.previous_page_number()
        else:
            prev_page_number = False
        page_ct = paginator.num_pages
        pages = [page_num]
        if paginator.num_pages > 1:  # all below creates abridged page list UI
            before = list(range(1, page_num))
            after = list(range(page_num + 1, paginator.num_pages + 1))
            default_ct = 2
            ct_before = (
                default_ct if len(after) > default_ct else default_ct * 2 - len(after)
            )
            ct_after = (
                default_ct if len(before) > default_ct else default_ct * 2 - len(before)
            )
            if len(before) > ct_before:
                before = [1, None] + before[-1 * (ct_before - 1) :]
            if len(after) > ct_after:
                after = after[0 : ct_after - 1] + [None, paginator.num_pages]
            pages = before + pages + after

        page_config = {
            "current_page": page_num,
            "end_index": page_results.end_index(),
            "has_next": page_results.has_next(),
            "has_other_pages": page_results.has_other_pages(),
            "has_previous": page_results.has_previous(),
            "next_page_number": next_page_number,
            "pages": pages,
            "previous_page_number": prev_page_number,
            "start_index": page_results.start_index(),
        }
        return JSONResponse(
            {"page_results": page_results.object_list, "paginator": page_config}
        )


def build_resource_dict(
    consultations, active_cons_node_list, datatype_factory, layout="grid", keyword=None
):
    """
    builds a list that looks like this:
    [
        {
            "resourceinstanceid":"[uuid],
            "node_name_a":"display_val_a",
            "node_name_b":"display_val_b",
            ...
        },
        {
            "resourceinstanceid":"[uuid],
            "node_name_a":"display_val_a",
            "node_name_b":"display_val_b",
            ...
        },
        ...
    ]
    """
    resources = []
    active_cons_list_vals = list(active_cons_node_list.values())
    active_cons_list_keys = list(active_cons_node_list.keys())
    for consultation in consultations:
        resource = {}
        resource["resourceinstanceid"] = consultation.resourceinstanceid
        consultation.load_tiles()
        for tile in consultation.tiles:
            for nodeid, nodevalue in list(tile.data.items()):
                if nodeid in active_cons_list_vals:
                    node = models.Node.objects.get(nodeid=nodeid)
                    try:
                        datatype = datatype_factory.get_instance(node.datatype)
                        val = datatype.get_display_value(tile, node)
                        if (
                            layout == "grid"
                            and nodeid
                            == active_cons_node_list["Geospatial Coordinates"]
                        ):
                            val = json.loads(val)
                    except Exception as e:
                        # print('Error:',e)
                        val = nodevalue

                    resource[str(node.name)] = val

        if keyword is not None:
            for v in list(resource.values()):
                try:
                    if keyword.lower() in v.lower():
                        resources.append(resource)
                        break
                except Exception as e:
                    continue
        else:
            for key in active_cons_list_keys:
                if key not in list(resource.keys()):
                    resource[key] = ""
            resources.append(resource)

    return resources



import uuid
from arches.app.functions.base import BaseFunction
from arches.app.models import models
from arches.app.models.tile import Tile
from arches.app.models.resource import Resource
from django.contrib.gis.geos import GEOSGeometry
import json
from datetime import datetime


details = {
    "name": "GeoJSON to BNG Point",
    "type": "node",
    "description": "Pushes the geometry from a GeoJSON node's centroid to a related BNG Point node",
    "defaultconfig": {
        "geojson_input_node": "",
        "bng_output_node": "",
        "geojson_input_nodegroup": "",
        "bng_output_nodegroup": "",
        "triggering_nodegroups": [],
    },
    "classname": "GeoJSONToBNGPoint",
    "component": "views/components/functions/geojson-to-bngpoint-function",
    "functionid": "d9a01773-6092-4cad-b331-ae725ae8fa88",
}


class GeoJSONToBNGPoint(BaseFunction):
    def get(self):
        raise NotImplementedError

    def save_bngpoint(self, tile, request, is_function_save_method=True):
        """Finds the BNG Alphanumeric value for the centroid of the envelope(extent) of the Geometry,
        and saves that value to the bng_output_nodegroup of the tile.

        Args:
            self: GeoJSONToBNGPoint object.

            tile: Tile to attach / amend bng_output_nodegroup of.

            request: WSGI Request used to varify call is result of user action. N.B. Function Returns if empty.

            is_function_save_method: a bool stating whether the function calling it is the save function.
        """

        # First let's check if this call is as a result of an inbound request (user action) or
        # as a result of the complementary BNGPointToGeoJSON function saving a new GeoJson.
        if request is None and is_function_save_method == True:
            return

        srid_LatLong = 4326
        srid_BngAbs = 27700
        # Reference grid for Easting/Northing to BNG Alphas.

        os_grid = {
            "09": "NA",
            "19": "NB",
            "29": "NC",
            "39": "ND",
            "49": "NE",
            "59": "OA",
            "69": "OB",
            "08": "NF",
            "18": "NG",
            "28": "NH",
            "38": "NJ",
            "48": "NK",
            "58": "OF",
            "68": "OG",
            "07": "NL",
            "17": "NM",
            "27": "NN",
            "37": "NO",
            "47": "NP",
            "57": "OL",
            "67": "OM",
            "06": "NQ",
            "16": "NR",
            "26": "NS",
            "36": "NT",
            "46": "NU",
            "56": "OQ",
            "66": "OR",
            "05": "NV",
            "15": "NW",
            "25": "NX",
            "35": "NY",
            "45": "NZ",
            "55": "OV",
            "65": "OW",
            "04": "SA",
            "14": "SB",
            "24": "SC",
            "34": "SD",
            "44": "SE",
            "54": "TA",
            "64": "TB",
            "03": "SF",
            "13": "SG",
            "23": "SH",
            "33": "SJ",
            "43": "SK",
            "53": "TF",
            "63": "TG",
            "02": "SL",
            "12": "SM",
            "22": "SN",
            "32": "SO",
            "42": "SP",
            "52": "TL",
            "62": "TM",
            "01": "SQ",
            "11": "SR",
            "21": "SS",
            "31": "ST",
            "41": "SU",
            "51": "TQ",
            "61": "TR",
            "00": "SV",
            "10": "SW",
            "20": "SX",
            "30": "SY",
            "40": "SZ",
            "50": "TV",
            "60": "TW",
        }

        geojsonnode = self.config["geojson_input_node"]
        bngnode = self.config["bng_output_node"]

        geojsonValue = tile.data[geojsonnode]

        if geojsonValue != None:

            # Grab a copy of the Geometry collection.
            geoJsFeatures = geojsonValue["features"]

            # Get the first feature as a GeosGeometry.
            geosGeom_union = GEOSGeometry(json.dumps(geoJsFeatures[0]["geometry"]))

            # update list.
            geoJsFeatures = geoJsFeatures[1:]

            # loop through list of geoJsFeatures.
            for item in geoJsFeatures:
                # .union seems to generate 'GEOS_ERROR: IllegalArgumentException:'
                # exceptions, but they seem spurious and are automatically ignored.
                geosGeom_union = geosGeom_union.union(GEOSGeometry(json.dumps(item["geometry"])))

            # find the centroid of the envelope for the resultant Geometry Collection.
            centroidPoint = geosGeom_union.envelope.centroid

            # Explicitly declare the SRID for the current lat/long.
            centroidPoint = GEOSGeometry(centroidPoint, srid=srid_LatLong)

            # Transform to Absolute BNG.
            centroidPoint.transform(srid_BngAbs, False)

            # Get initial Easting and Northing digits. N.B. Left Zero pad integer coords to 6 digits!
            easting = str(int(centroidPoint.coords[0])).zfill(6)
            northing = str(int(centroidPoint.coords[1])).zfill(6)
            gridref = easting[0] + northing[0]

            # Get AlphaNumeric BNG
            try:
                gridref = os_grid[gridref] + easting[1:6] + northing[1:6]
            except KeyError:
                raise Exception("Conversion Error : Coordinates outside of BNG for England.")

            if self.config["bng_output_nodegroup"] == str(tile.nodegroup_id):
                tile.data[bngnode] = gridref
            else:

                previously_saved_tiles = Tile.objects.filter(
                    nodegroup_id=self.config["bng_output_nodegroup"], resourceinstance_id=tile.resourceinstance_id
                )

                # Update pre-existing tiles, or Create new one.
                if len(previously_saved_tiles) > 0:
                    for p in previously_saved_tiles:
                        p.data[bngnode] = gridref
                        p.save()
                else:
                    new_bng_tile = Tile().get_blank_tile_from_nodegroup_id(
                        self.config["bng_output_nodegroup"], resourceid=tile.resourceinstance_id, parenttile=tile.parenttile
                    )
                    new_bng_tile.data[bngnode] = gridref
                    new_bng_tile.save()

            return

        return

    def save(self, tile, request, context=None):

        self.save_bngpoint(tile=tile, request=request, is_function_save_method=True)
        return

    def post_save(self, *args, **kwargs):
        raise NotImplementedError

    def delete(self, tile, request):
        raise NotImplementedError

    def on_import(self, tile):
        self.save_bngpoint(tile=tile, request=None, is_function_save_method=False)
        return

    def after_function_save(self, tile, request):
        raise NotImplementedError
